"""
Different functions for extracting the fields and terminal currents and voltages from the solution vector.

Extract the different fields.
Extract the losses and energy.
Extract the integral quantities.
Extract the source terminal currents and voltages.
Compute the magnetic field for the point cloud.

Warning
-------
    - The magnetic field computation is done with lumped variables.
    - Therefore, the computation is only accurate far away from the voxel structure.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np
import numpy.linalg as lna
from pypeec import log

# get a logger
LOGGER = log.get_logger("SOLUTION")


def _get_biot_savart(pts, pts_net, J_src, vol):
    """
    Compute the magnetic field at a specified point.
    The field is created by many currents.
    """

    # get the distance between the points and the voxels
    vec = pts-pts_net

    # get the norm of the distance
    nrm = lna.norm(vec, axis=1, keepdims=True)

    # compute the Biot-Savart contributions
    H_all = (vol/(4*np.pi))*(np.cross(J_src, vec, axis=1)/(nrm**3))

    # sum the contributions
    H_pts = np.sum(H_all, axis=0)

    return H_pts


def _get_magnetic_charge(pts, pts_src, Q_src, vol):
    """
    Compute the magnetic field at a specified point.
    The field is created by many magnetic charge.
    """

    # vacuum permeability
    mu = 4*np.pi*1e-7

    # get the distance between the points and the voxels
    vec = pts_src-pts

    # get the norm of the distance
    nrm = lna.norm(vec, axis=1, keepdims=True)

    # transform the charge into a vector
    Q_src = np.tile(Q_src, (3, 1)).transpose()

    # compute the Biot-Savart contributions
    H_all = (vol/(4*np.pi*mu))*((Q_src*vec)/(nrm**3))

    # sum the contributions
    H_pts = np.sum(H_all, axis=0)

    return H_pts


def get_magnetic_field(d, J_vc, Q_vm, pts_net_c, pts_net_m, pts_cloud):
    """
    Compute the magnetic field for the provided points.
    The Biot-Savart law is used for the electric material contribution.
    The magnetic charge is used for the magnetic material contribution.
    """

    # extract the voxel volume
    vol = np.prod(d)

    # for each provided point, compute the magnetic field
    H_pts = np.zeros((len(pts_cloud), 3), dtype=np.complex_)
    for i, pts_tmp in enumerate(pts_cloud):
        H_c = _get_biot_savart(pts_tmp, pts_net_c, J_vc, vol)
        H_m = _get_magnetic_charge(pts_tmp, pts_net_m, Q_vm, vol)
        H_pts[i, :] = H_c+H_m

    return H_pts


def get_sol_extract(sol, sol_idx):
    """
    Extract the different variables from the solution vector.
    """

    I_fc = sol[sol_idx["I_fc"]]
    V_vc = sol[sol_idx["V_vc"]]
    I_fm = sol[sol_idx["I_fm"]]
    V_vm = sol[sol_idx["V_vm"]]
    I_src = sol[sol_idx["I_src"]]

    return I_fc, V_vc, I_fm, V_vm, I_src


def get_vector_density(n, d, idx_f, A_net, var_f):
    """
    Project a face vector variable into a voxel vector variable.
    Scale the variable with respect to the face area (density).

    At the input, the array has the following size: 3*nx*ny*nx.
    At the output, the array has the following size: (nx*ny*nx, 3).
    """

    # extract the voxel data
    (nx, ny, nz) = n
    nv = nx*ny*nz

    # extract the voxel data
    (dx, dy, dz) = d

    # get the direction of the faces (x, y, z)
    idx_fx = np.in1d(idx_f, np.arange(0*nv, 1*nv, dtype=np.int_))
    idx_fy = np.in1d(idx_f, np.arange(1*nv, 2*nv, dtype=np.int_))
    idx_fz = np.in1d(idx_f, np.arange(2*nv, 3*nv, dtype=np.int_))

    # project the faces into the voxels
    var_v_x = 0.5*np.abs(A_net[:, idx_fx])*var_f[idx_fx]
    var_v_y = 0.5*np.abs(A_net[:, idx_fy])*var_f[idx_fy]
    var_v_z = 0.5*np.abs(A_net[:, idx_fz])*var_f[idx_fz]

    # convert to density.
    var_v_x = var_v_x/(dy*dz)
    var_v_y = var_v_y/(dx*dz)
    var_v_z = var_v_z/(dx*dy)

    # assemble the variables
    var_v = np.stack((var_v_x, var_v_y, var_v_z), axis=1)

    return var_v


def get_scalar_density(d, A_net, var_f):
    """
    Project a face vector variable into a voxel scalar variable.
    Scale the variable with respect to the voxel volume (density).

    At the input, the array has the following size: 3*nx*ny*nx.
    At the output, the array has the following size: nx*ny*nx.
    """

    # extract the voxel data
    (dx, dy, dz) = d

    # convert face to voxel variable
    var_v = 0.5*np.abs(A_net)*var_f

    # convert to density.
    var_v = var_v/(dx*dy*dz)

    return var_v


def get_divergence_density(d, A_net, var_f):
    """
    Compute the divergence of a face vector with respect to the voxels.
    Scale the variable with respect to the voxel volume (density).

    At the input, the array has the following size: 3*nx*ny*nx.
    At the output, the array has the following size: nx*ny*nx.
    """

    # extract the voxel data
    (dx, dy, dz) = d

    # compute the divergence
    var_v = A_net*var_f

    # convert to density.
    var_v = var_v/(dx*dy*dz)

    return var_v


def get_losses(freq, I_fc, I_fm, R_c, R_m):
    """
    Get the losses for the electric and magnetic domains.
    """

    # get the angular frequency
    s = 1j*2*np.pi*freq

    # get the factor for getting the loss time-averaged values
    if freq == 0:
        fact = 1.0
    else:
        fact = 0.5

    # get the magnetic losses linked with the electric domains
    P_fc = fact*np.conj(I_fc)*R_c*I_fc
    P_fc = np.real(P_fc)

    # get the magnetic losses linked with the magnetic domains
    P_fm = fact*np.conj(s*I_fm)*R_m*I_fm
    P_fm = np.real(P_fm)

    return P_fc, P_fm


def get_energy(freq, I_fc, I_fm, L_op_c, K_op_c):
    """
    Get the energy for the electric and magnetic domains.
    """

    # get the factor for getting the energy time-averaged values
    if freq == 0:
        fact = 0.5
    else:
        fact = 0.25

    # get the magnetic energy linked with the electric domains
    W_fc = fact*np.conj(I_fc)*L_op_c(I_fc)
    W_fc = np.real(W_fc)

    # get the magnetic energy linked with the magnetic domains
    W_fm = fact*np.conj(I_fc)*K_op_c(I_fm)
    W_fm = np.real(W_fm)

    return W_fc, W_fm


def get_integral(P_fc, P_fm, W_fc, W_fm, S_tot):
    """
    Sum the loss/energy in order to obtain global quantities.
    """

    # compute the integral quantities
    P_electric = np.sum(P_fc)
    P_magnetic = np.sum(P_fm)
    W_electric = np.sum(W_fc)
    W_magnetic = np.sum(W_fm)
    P_tot = P_electric+P_magnetic
    W_tot = W_electric+W_magnetic

    # assign the integral quantities
    integral = {
        "P_electric": P_electric, "P_magnetic": P_magnetic,
        "W_electric": W_electric, "W_magnetic": W_magnetic,
        "P_tot": P_tot, "W_tot": W_tot, "S_tot": S_tot,
    }

    # display
    LOGGER.debug("integral")
    with log.BlockIndent():
        LOGGER.debug("S_tot_real = %.2e VA" % S_tot.real)
        LOGGER.debug("S_tot_imag = %.2ej VA" % S_tot.imag)
        LOGGER.debug("P_electric = %.2e W" % P_electric)
        LOGGER.debug("P_magnetic = %.2e W" % P_magnetic)
        LOGGER.debug("W_electric = %.2e J" % W_electric)
        LOGGER.debug("W_magnetic = %.2e J" % W_magnetic)
        LOGGER.debug("P_tot = %.2e W" % P_tot)
        LOGGER.debug("W_tot = %.2e J" % W_tot)

    return integral


def get_material(material_pos, A_net_c, A_net_m, P_fc, P_fm):
    """
    Parse the losses for the materials.
    The results are assigned to a dict with the magnetic and electric losses.
    """

    # init material dict
    material = {}

    # compute the losses of the voxels
    P_vc = 0.5*np.abs(A_net_c)*P_fc
    P_vm = 0.5*np.abs(A_net_m)*P_fm

    # parse the material domains
    for tag, material_pos_tmp in material_pos.items():
        # extract the data
        idx_vc = material_pos_tmp["idx_vc"]
        idx_vm = material_pos_tmp["idx_vm"]

        # get the domain losses
        P_vc_tmp = np.sum(P_vc[idx_vc])
        P_vm_tmp = np.sum(P_vm[idx_vm])
        P_tmp = P_vc_tmp+P_vm_tmp

        # assign the losses
        material[tag] = {"P_electric": P_vc_tmp, "P_magnetic": P_vm_tmp, "P_tot": P_tmp}

        # display
        LOGGER.debug("domain: %s" % tag)
        with log.BlockIndent():
            LOGGER.debug("P_electric = %.2e W" % P_vc_tmp)
            LOGGER.debug("P_magnetic = %.2e W" % P_vm_tmp)
            LOGGER.debug("P_tot = %.2e W" % P_tmp)

    return material


def get_source(freq, source_all, I_src, V_vc):
    """
    Parse the terminal voltages and currents for the sources.
    The sources have internal resistances/admittances.
    Therefore, the extracted value can differ from the prescribed value.
    The results are assigned to a dict with the voltage, current, and power values.
    """

    # init source dict
    source = {}

    # total complex power
    S_tot = 0.0

    # get the factor for getting the power time-averaged values
    if freq == 0:
        fact = 1.0
    else:
        fact = 0.5

    # parse the source terminals
    for tag, source_all_tmp in source_all.items():
        # extract the data
        idx = source_all_tmp["idx"]
        idx_vc = source_all_tmp["idx_vc"]
        idx_src = source_all_tmp["idx_src"]
        source_type = source_all_tmp["source_type"]
        var_type = source_all_tmp["var_type"]
        value = source_all_tmp["value"]
        element = source_all_tmp["element"]

        # get the distributed source
        if len(idx) == 0:
            V_tmp = np.complex_(0)
            I_tmp = np.complex_(0)
            value = np.complex_(0)
            element = np.complex_(0)
        else:
            V_tmp = np.complex_(V_vc[idx_vc])
            I_tmp = np.complex_(I_src[idx_src])

        # get the drop across the source impedance
        if source_type == "current":
            drop_tmp = np.sum(V_tmp*element)
            src_tmp = np.sum(value)
        elif source_type == "voltage":
            drop_tmp = np.mean(I_tmp*element)
            src_tmp = np.mean(value)
        else:
            raise ValueError("invalid source type")

        # compute the lumped quantities
        S_tmp = np.sum(fact*V_tmp*np.conj(I_tmp))
        V_tmp = np.mean(V_tmp)
        I_tmp = np.sum(I_tmp)

        # assign the current and voltage
        source[tag] = {
            "V": V_tmp, "I": I_tmp, "S": S_tmp,
            "source_type": source_type, "var_type": var_type,
            "src": src_tmp, "drop": drop_tmp,
        }

        # add the power
        S_tot += S_tmp

        # display
        LOGGER.debug("terminal: %s" % tag)
        with log.BlockIndent():
            # source type
            LOGGER.debug("type = %s / %s" % (source_type, var_type))

            # source value
            if source_type == "current":
                LOGGER.debug("I_src = %+.2e + %+.2ej V" % (src_tmp.real, src_tmp.imag))
                LOGGER.debug("I_drop = %+.2e + %+.2ej V" % (drop_tmp.real, drop_tmp.imag))
            elif source_type == "voltage":
                LOGGER.debug("V_src = %+.2e + %+.2ej V" % (src_tmp.real, src_tmp.imag))
                LOGGER.debug("V_drop = %+.2e + %+.2ej V" % (drop_tmp.real, drop_tmp.imag))
            else:
                raise ValueError("invalid source type")

            # terminal value
            LOGGER.debug("V = %+.2e + %+.2ej V" % (V_tmp.real, V_tmp.imag))
            LOGGER.debug("I = %+.2e + %+.2ej A" % (I_tmp.real, I_tmp.imag))
            LOGGER.debug("S = %+.2e + %+.2ej VA" % (S_tmp.real, S_tmp.imag))

    return source, S_tot



"""
Set the package metadata.
"""

import importlib.resources

# set basic metadata
__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

# get the version number
try:
    with importlib.resources.open_text("pypeec.data", "version.txt") as file_version:
        __version__ = file_version.read()
except FileNotFoundError:
    __version__ = 'x.x.x'



"""
Provide a function to extract convergence metrics from the solution vector.

The complex power is extracted at the different terminals.
The total complex power is computed for all the terminals.
The active and reactive power are returned as metrics.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np


def _get_sol_extract(sol, sol_idx):
    """
    Extract the different variables from the solution vector.
    """

    V_vc = sol[sol_idx["V_vc"]]
    I_src = sol[sol_idx["I_src"]]

    return V_vc, I_src


def _get_total_power(freq, source_all, I_src, V_vc):
    """
    Compute the total complex power is computed for all the terminals.
    """

    # get the factor for getting the power time-averaged values
    if freq == 0:
        fact = 1.0
    else:
        fact = 0.5

    # init terminal power
    power = 0.0

    # parse the source terminals
    for tag, source_all_tmp in source_all.items():
        # extract the data
        idx_vc = source_all_tmp["idx_vc"]
        idx_src = source_all_tmp["idx_src"]

        # get the current and voltage
        if (len(idx_vc) != 0) and (len(idx_src) != 0):
            V_tmp = V_vc[idx_vc]
            I_tmp = I_src[idx_src]

            # compute the apparent power
            power_tmp = np.sum(fact*V_tmp*np.conj(I_tmp))

            # add the power
            power += power_tmp

    return power


def _get_conv_eval(sol, freq, source_all, sol_idx):
    """
    Extract the convergence metrics (active and reactive power) from a solution vector.
    """

    # extract the data
    (V_vc, I_src) = _get_sol_extract(sol, sol_idx)

    # get the sources
    power = _get_total_power(freq, source_all, I_src, V_vc)

    return power


def get_fct_conv(freq, source_all, sol_idx):
    """
    Return a function that extract convergence metrics from the solution vector.
    """

    def fct_conv(sol):
        power = _get_conv_eval(sol, freq, source_all, sol_idx)
        return power

    return fct_conv



"""
Main script for solving a problem with the PEEC solver.
Check the input data, solve the problem, and parse the results.
The different parts of the code are timed.

The solver is implemented with NumPy and Scipy.
Optional HPC libraries can be used to accelerate the solver.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

from pypeec.lib_matrix import matrix_factorization
from pypeec.lib_matrix import multiply_fft
from pypeec.lib_matrix import fourier_transform
from pypeec.lib_solver import sweep_solver
from pypeec.lib_solver import voxel_geometry
from pypeec.lib_solver import system_tensor
from pypeec.lib_solver import problem_geometry
from pypeec.lib_solver import problem_value
from pypeec.lib_solver import system_matrix
from pypeec.lib_solver import equation_system
from pypeec.lib_solver import equation_solver
from pypeec.lib_solver import extract_solution
from pypeec.lib_solver import extract_convergence
from pypeec.lib_check import check_data_problem
from pypeec.lib_check import check_data_tolerance
from pypeec.lib_check import check_data_solver
from pypeec.lib_check import check_data_options
from pypeec import log


# get a logger
LOGGER = log.get_logger("SOLVER")


def _run_solver_options(data_solver):
    """
    Load and configure the optional libraries.
    """

    # extract the data
    factorization_options = data_solver["factorization_options"]
    fft_options = data_solver["fft_options"]

    # set options
    matrix_factorization.set_options(factorization_options)
    fourier_transform.set_options(fft_options)
    multiply_fft.set_options(fft_options)


def _run_solver_init(data_solver, is_truncated):
    """
    Compute the voxel geometry, Green functions, and the incidence matrix.
    """

    # extract the data
    n = data_solver["n"]
    d = data_solver["d"]
    c = data_solver["c"]
    green_simplify = data_solver["green_simplify"]
    coupling_simplify = data_solver["coupling_simplify"]
    mult_type = data_solver["mult_type"]
    has_coupling = data_solver["has_coupling"]
    has_electric = data_solver["has_electric"]
    has_magnetic = data_solver["has_magnetic"]
    material_idx = data_solver["material_idx"]
    source_idx = data_solver["source_idx"]
    pts_cloud = data_solver["pts_cloud"]

    # load and configure the optional libraries
    _run_solver_options(data_solver)

    # get the voxel geometry and the incidence matrix
    with log.BlockTimer(LOGGER, "voxel_geometry"):
        # get the coordinate of the voxels
        pts_vox = voxel_geometry.get_voxel_coordinate(n, d, c)

        # compute the incidence matrix
        A_vox = voxel_geometry.get_incidence_matrix(n)

    # get the Green functions
    with log.BlockTimer(LOGGER, "system_tensor"):
        # Green function self-coefficient
        G_self = system_tensor.get_green_self(d)

        # Green function mutual coefficients
        G_mutual = system_tensor.get_green_tensor(n, d, green_simplify)

        # Green function mutual coefficients
        K_tsr = system_tensor.get_coupling_tensor(n, d, coupling_simplify, has_coupling)

    # parse the problem geometry (materials and sources)
    with log.BlockTimer(LOGGER, "problem_geometry"):
        # parse the materials
        (idx_vc, idx_vm) = problem_geometry.get_material_indices(material_idx)
        material_pos = problem_geometry.get_material_pos(material_idx, idx_vc, idx_vm)

        # parse the sources
        (idx_src_c, idx_src_v) = problem_geometry.get_source_indices(source_idx)
        source_pos = problem_geometry.get_source_pos(source_idx, idx_vc, idx_src_c, idx_src_v)

        # reduce the incidence matrix to the non-empty voxels and compute face indices
        (pts_net_c, A_net_c, idx_fc) = problem_geometry.get_reduce_matrix(pts_vox, A_vox, idx_vc)
        (pts_net_m, A_net_m, idx_fm) = problem_geometry.get_reduce_matrix(pts_vox, A_vox, idx_vm)

        # free memory
        del pts_vox
        del A_vox

        # get a summary of the problem size
        problem_status = problem_geometry.get_status(n, idx_vc, idx_vm, idx_fc, idx_fm, idx_src_c, idx_src_v)

    # get the system operators
    with log.BlockTimer(LOGGER, "system_matrix"):
        # get the inductance tensor (preconditioner and full problem)
        (L_c, L_op_c) = system_matrix.get_inductance_matrix(n, d, idx_fc, G_self, G_mutual, has_electric, mult_type)

        # get the potential tensor (preconditioner and full problem)
        (P_m, P_op_m) = system_matrix.get_potential_matrix(d, idx_vm, G_self, G_mutual, has_magnetic, mult_type)

        # free memory
        del G_self
        del G_mutual

        # get the coupling matrices
        (K_op_c, K_op_m) = system_matrix.get_coupling_matrix(n, idx_vc, idx_vm, idx_fc, idx_fm, A_net_c, A_net_m, K_tsr, has_coupling, mult_type)

        # free memory
        del K_tsr

    # assign the results (internal data required to solve the problem)
    data_internal = {
        "idx_vc": idx_vc,
        "idx_vm": idx_vm,
        "idx_fc": idx_fc,
        "idx_fm": idx_fm,
        "idx_src_c": idx_src_c,
        "idx_src_v": idx_src_v,
        "A_net_c": A_net_c,
        "A_net_m": A_net_m,
        "L_c": L_c,
        "L_op_c": L_op_c,
        "P_m": P_m,
        "P_op_m": P_op_m,
        "K_op_c": K_op_c,
        "K_op_m": K_op_m,
        "source_pos": source_pos,
        "material_pos": material_pos,
        "pts_net_c": pts_net_c,
        "pts_net_m": pts_net_m,
    }

    # assign the results (will be merged in the solver output)
    data_init = {
        "n": n,
        "d": d,
        "c": c,
        "problem_status": problem_status,
    }

    # if required, add the complete data
    if not is_truncated:
        data_add = {
            "idx_vc": idx_vc,
            "idx_vm": idx_vm,
            "idx_fc": idx_fc,
            "idx_fm": idx_fm,
            "idx_src_c": idx_src_c,
            "idx_src_v": idx_src_v,
            "pts_cloud": pts_cloud,
        }
        data_init = {**data_init, **data_add}

    return data_init, data_internal


def _run_solver_sweep(data_solver, data_internal, sweep_param, sol_init, is_truncated):
    """
    Create the equation system, solve the system, and extract the solution.
    """

    # extract the data
    n = data_solver["n"]
    d = data_solver["d"]
    material_idx = data_solver["material_idx"]
    source_idx = data_solver["source_idx"]
    condition_options = data_solver["condition_options"]
    solver_options = data_solver["solver_options"]
    pts_cloud = data_solver["pts_cloud"]

    # extract the data
    idx_vc = data_internal["idx_vc"]
    idx_vm = data_internal["idx_vm"]
    idx_fc = data_internal["idx_fc"]
    idx_fm = data_internal["idx_fm"]
    idx_src_c = data_internal["idx_src_c"]
    idx_src_v = data_internal["idx_src_v"]
    A_net_c = data_internal["A_net_c"]
    A_net_m = data_internal["A_net_m"]
    L_c = data_internal["L_c"]
    L_op_c = data_internal["L_op_c"]
    P_m = data_internal["P_m"]
    P_op_m = data_internal["P_op_m"]
    K_op_c = data_internal["K_op_c"]
    K_op_m = data_internal["K_op_m"]
    material_pos = data_internal["material_pos"]
    source_pos = data_internal["source_pos"]
    pts_net_c = data_internal["pts_net_c"]
    pts_net_m = data_internal["pts_net_m"]

    # extract the data
    freq = sweep_param["freq"]
    material_val = sweep_param["material_val"]
    source_val = sweep_param["source_val"]

    # load and configure the optional libraries
    _run_solver_options(data_solver)

    # get the material and source values
    with log.BlockTimer(LOGGER, "problem_value"):
        # parse the material parameters
        (rho_vc, rho_vm) = problem_value.get_material_vector(material_val, material_idx)

        # parse the source parameters
        source_all = problem_value.get_source_all(source_val, source_pos, source_idx)
        (I_src_c, G_src_c) = problem_value.get_source_vector(source_all, "current")
        (V_src_v, R_src_v) = problem_value.get_source_vector(source_all, "voltage")

        # get the resistance vector
        R_c = problem_value.get_resistance_vector(n, d, A_net_c, idx_fc, rho_vc)
        R_m = problem_value.get_resistance_vector(n, d, A_net_m, idx_fm, rho_vm)

    # assemble the equation system
    with log.BlockTimer(LOGGER, "equation_system"):
        # compute the right-hand vector with the sources
        rhs = equation_system.get_source_vector(idx_vc, idx_vm, idx_fc, idx_fm, I_src_c, V_src_v)

        # get the source connection matrices
        A_src = equation_system.get_source_matrix(idx_vc, idx_src_c, idx_src_v, G_src_c, R_src_v)

        # get the linear operator for the preconditioner (guess of the inverse)
        (fct_pcd, S_mat_c, S_mat_m) = equation_system.get_cond_operator(freq, A_net_c, A_net_m, A_src, R_c, R_m, L_c, P_m)

        # get the linear operator for the electric-magnetic coupling
        fct_cpl = equation_system.get_coupling_operator(freq, A_net_c, A_net_m, A_src, K_op_c, K_op_m)

        # get the linear operator for the full system (matrix-vector multiplication)
        fct_sys = equation_system.get_system_operator(freq, A_net_c, A_net_m, A_src, R_c, R_m, L_op_c, P_op_m)

        # get the source indices
        sol_idx = equation_system.get_system_sol_idx(A_net_c, A_net_m, A_src)

    # solve the equation system
    with log.BlockTimer(LOGGER, "equation_solver"):
        # estimate the condition number of the problem (to detect quasi-singular problem)
        (condition_ok, condition_status) = equation_solver.get_condition(S_mat_c, S_mat_m, condition_options)

        # free memory
        del S_mat_c
        del S_mat_m

        # get a function to evaluate the solver convergence
        fct_conv = extract_convergence.get_fct_conv(freq, source_all, sol_idx)

        # solve the equation system
        (sol, res, conv, solver_ok, solver_status) = equation_solver.get_solver(sol_init, fct_cpl, fct_sys, fct_pcd, rhs, fct_conv, solver_options)

        # free memory
        del fct_pcd
        del fct_cpl
        del fct_sys

        # compute convergence
        has_converged = solver_ok and condition_ok

    # extract the solution
    with log.BlockTimer(LOGGER, "extract_solution"):
        # split the solution vector to get the face currents, the voxel potentials, and the sources
        (I_fc, V_vc, I_fm, V_vm, I_src) = extract_solution.get_sol_extract(sol, sol_idx)

        # get the losses and energy
        (P_fc, P_fm) = extract_solution.get_losses(freq, I_fc, I_fm, R_c, R_m)
        (W_fc, W_fm) = extract_solution.get_energy(freq, I_fc, I_fm, L_op_c, K_op_c)

        # get the voxel flow densities from the face flows
        J_vc = extract_solution.get_vector_density(n, d, idx_fc, A_net_c, I_fc)
        B_vm = extract_solution.get_vector_density(n, d, idx_fm, A_net_m, I_fm)

        # get the voxel loss densities from the face losses
        P_vc = extract_solution.get_scalar_density(d, A_net_c, P_fc)
        P_vm = extract_solution.get_scalar_density(d, A_net_m, P_fm)

        # get the divergence of the face flows
        S_vc = extract_solution.get_divergence_density(d, A_net_c, I_fc)
        Q_vm = extract_solution.get_divergence_density(d, A_net_m, I_fm)

        # get the domain losses for the different materials
        material = extract_solution.get_material(material_pos, A_net_c, A_net_m, P_fc, P_fm)

        # get the terminal voltages and currents for the sources
        (source, S_tot) = extract_solution.get_source(freq, source_all, I_src, V_vc)

        # get the global quantities (energy and losses)
        integral = extract_solution.get_integral(P_fc, P_fm, W_fc, W_fm, S_tot)

        # get the cloud point magnetic field
        H_pts = extract_solution.get_magnetic_field(d, J_vc, Q_vm, pts_net_c, pts_net_m, pts_cloud)

    # assign the results (will be merged in the solver output)
    data_sweep = {
        "freq": freq,
        "has_converged": has_converged,
        "solver_status": solver_status,
        "condition_status": condition_status,
        "material": material,
        "source": source,
        "integral": integral,
        "conv": conv,
    }

    # if required, add the complete data
    if not is_truncated:
        data_add = {
            "res": res,
            "V_vc": V_vc,
            "V_vm": V_vm,
            "J_vc": J_vc,
            "B_vm": B_vm,
            "P_vc": P_vc,
            "P_vm": P_vm,
            "S_vc": S_vc,
            "Q_vm": Q_vm,
            "H_pts": H_pts,
        }
        data_sweep = {**data_sweep, **data_add}

    return data_sweep, sol


def _get_data(data_init, data_sweep, timestamp, is_truncated):
    """
    Assemble the returned data.
    """

    # end message
    (duration, fmt) = log.get_duration(timestamp)
    LOGGER.info("duration: %s" % fmt)

    # cast to seconds
    duration = duration.total_seconds()

    # extract the solution
    data_solution = {
        "duration": duration,
        "is_truncated": is_truncated,
        "data_init": data_init,
        "data_sweep": data_sweep,
    }

    return data_solution


def run(data_voxel, data_problem, data_tolerance, is_truncated=False):
    """
    Main script for solving a problem with the PEEC solver.
    Handle invalid data with exceptions.
    """

    # get timestamp
    timestamp = log.get_timer()

    # check the voxel data
    LOGGER.info("check the voxel data")
    data_geom = check_data_options.check_data_voxel(data_voxel)

    # check the input data
    LOGGER.info("check the input data")
    check_data_problem.check_data_problem(data_problem)
    check_data_tolerance.check_data_tolerance(data_tolerance)
    check_data_options.check_data_options(is_truncated)

    # combine the problem and voxel data
    LOGGER.info("combine the input data")
    (data_solver, sweep_config, sweep_param) = check_data_solver.get_data_solver(data_geom, data_problem, data_tolerance)

    # create the problem
    with log.BlockTimer(LOGGER, "init"):
        (data_init, data_internal) = _run_solver_init(data_solver, is_truncated)

    # function for solving a single sweep
    def fct_compute(tag, data, init):
        with log.BlockTimer(LOGGER, "run sweep: " + tag):
            (output, init) = _run_solver_sweep(data_solver, data_internal, data, init, is_truncated)
        return output, init

    # compute the different sweeps
    data_sweep = sweep_solver.get_run_sweep(sweep_config, sweep_param, fct_compute)

    # create output data
    data_solution = _get_data(data_init, data_sweep, timestamp, is_truncated)

    return data_solution



"""
Module for generating PNG files from GERBER files.

The PCB is created with "KiCad - PCB" and the following files are exported:
    - GERBER files for the different layers (main layers)
    - GERBER files for the terminals (user layers)
    - Excellon files for the vias

Finally, this script is used to generate PNG files for the layers:
    - using "gerbv - Gerber Viewer 2.8" for parsing the GERBER files
    - using "mogrify - ImageMagick 6.9" for trimming the PNG files

Warning
-------
    - This code requires the following external tools: "gerbv" and "mogrify".
    - This code has only been tested on Linux (Ubuntu 22.04).
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import os
import os.path
import subprocess


def _get_gerbv_file(filename_gerbv, folder_gerber, data_gerber, data_stack):
    """
    Create a "gerbv - Gerber Viewer" project file.
    """

    # extract the data
    gerber_edge = data_gerber["gerber_edge"]
    color_edge = data_gerber["color_edge"]
    color_background = data_gerber["color_background"]
    color_def = data_gerber["color_def"]
    gerber_def = data_gerber["gerber_def"]

    # get base color
    color_edge = tuple([257 * x for x in color_edge])
    color_background = tuple([257 * x for x in color_background])

    # get GERBER edge file
    gerber_edge = os.path.join(folder_gerber, gerber_edge)

    # get transparency value
    alpha_channel = 257*255

    # create the file
    with open(filename_gerbv, "w") as fid:
        # add header
        fid.write('(gerbv-file-version! "2.0A")\n')

        # add stack
        for i, data_stack_tmp in enumerate(data_stack):
            # extract the data
            layer = i+1
            gerber = data_stack_tmp["gerber"]
            color = data_stack_tmp["color"]

            # get layer color
            color = color_def[color]
            color = tuple([257*x for x in color])

            # get GERBER file
            gerber = gerber_def[gerber]
            gerber = os.path.join(folder_gerber, gerber)

            # add layer
            fid.write('(define-layer! %d (cons \'filename "%s")\n' % (layer, gerber))
            fid.write('\t(cons \'visible #t)\n')
            fid.write('\t(cons \'color #(%d %d %d))\n' % color)
            fid.write('\t(cons \'alpha #(%d))\n' % alpha_channel)
            fid.write(')\n')

        # add board edges
        fid.write('(define-layer! 0 (cons \'filename "%s")\n' % gerber_edge)
        fid.write('\t(cons \'visible #t)\n')
        fid.write('\t(cons \'color #(%d %d %d))\n' % tuple(color_edge))
        fid.write('\t(cons \'alpha #(%d))\n' % alpha_channel)
        fid.write(')\n')

        # add background
        fid.write('(define-layer! -1 (cons \'filename "%s")\n' % filename_gerbv)
        fid.write('\t(cons \'color #(%d %d %d))\n' % tuple(color_background))
        fid.write('\t(cons \'alpha #(%d))\n' % alpha_channel)
        fid.write(')\n')

        # add rendering settings
        fid.write('(set-render-type! 3)\n')


def _get_png_file(filename_gerbv, filename_png, margin, voxel, oversampling):
    """
    Transform GERBER project file into a PNG file for a given layer.
    """

    # get the resolution and margin
    resolution = round(25.4e-3*oversampling/voxel)
    resize = 100/oversampling
    border = 100*margin

    # transform the GERBER files into a PNG file
    cmd = "gerbv -p %s -o %s -x png --border %d --dpi %d 2>/dev/null" % (
        filename_gerbv,
        filename_png,
        border,
        resolution,
    )
    subprocess.run(cmd, shell=True)

    # cut the border
    cmd = "mogrify -trim %s" % filename_png
    subprocess.run(cmd, shell=True)

    # cut the border
    cmd = "mogrify -sample %.2f%% %s" % (resize, filename_png)
    subprocess.run(cmd, shell=True)


def _get_convert_layer(tag, data_export, data_gerber, data_stack):
    """
    Transform GERBER files into a PNG file for a given layer.
    """

    # extract the data
    margin = data_export["margin"]
    voxel = data_export["voxel"]
    oversampling = data_export["oversampling"]
    folder_gerber = data_export["folder_gerber"]
    folder_png = data_export["folder_png"]

    # get the absolute location
    folder_gerber = os.path.abspath(folder_gerber)
    folder_png = os.path.abspath(folder_png)

    # create the file names
    filename_gerbv = os.path.join(folder_png, "%s.gvp" % tag)
    filename_png = os.path.join(folder_png, "%s.png" % tag)

    # get the project file with the GERBER files
    _get_gerbv_file(filename_gerbv, folder_gerber, data_gerber, data_stack)

    # convert the project file into a PNG file
    _get_png_file(filename_gerbv, filename_png, margin, voxel, oversampling)

    # remove the project file
    os.remove(filename_gerbv)


def get_convert(data_export, data_gerber, data_stack):
    """
    Transform GERBER files into PNG files.
    """

    for tag, data_stack_tmp in data_stack.items():
        _get_convert_layer(tag, data_export, data_gerber, data_stack_tmp)



"""
Module for computer FFT/iFFT of tensors.

This module is used as a common interface for different FFT libraries:
    - SciPy FFT library
    - FFTW FFT library (available through pyFFTW)
    - CuPy FFT library (computation with GPUs)
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import os

# dummy functions
FFTN = None
IFFTN = None


def set_options(fft_options):
    """
    Assign the options and load the right libray.
    """

    # assign global variable
    library = fft_options["library"]
    scipy_worker = fft_options["scipy_worker"]
    fftw_thread = fft_options["fftw_thread"]
    fftw_timeout = fft_options["fftw_timeout"]
    fftw_byte_align = fft_options["fftw_byte_align"]

    # import the right library
    if library == "CuPy":
        import cupy.fft as fftc

        # find FFTN function
        def fct_fftn(mat, shape, axes, _):
            return fftc.fftn(mat, shape, axes=axes)

        # find iFFTN function
        def fct_ifftn(mat, shape, axes, _):
            return fftc.ifftn(mat, shape, axes=axes)
    elif library == "SciPy":
        import scipy.fft as ffts

        # find the number of workers
        if scipy_worker is None:
            scipy_worker = os.cpu_count()

        # find FFTN function
        def fct_fftn(mat, shape, axes, replace):
            return ffts.fftn(mat, shape, axes=axes, overwrite_x=replace, workers=scipy_worker)

        # find iFFTN function
        def fct_ifftn(mat, shape, axes, replace):
            return ffts.ifftn(mat, shape, axes=axes, overwrite_x=replace, workers=scipy_worker)
    elif library == "FFTW":
        import pyfftw
        from pyfftw.interfaces import cache
        from pyfftw.interfaces import numpy_fft

        # find the number of threads
        if fftw_thread is None:
            fftw_thread = os.cpu_count()

        # configure the FFT cache
        if fftw_timeout is None:
            cache.disable()
        else:
            cache.enable()
            cache.set_keepalive_time(fftw_timeout)

        # find FFTN function
        def fct_fftn(mat, shape, axes, replace):
            mat = pyfftw.byte_align(mat, n=fftw_byte_align)
            return numpy_fft.fftn(mat, shape, axes=axes, overwrite_input=replace, threads=fftw_thread)

        # find iFFTN function
        def fct_ifftn(mat, shape, axes, replace):
            mat = pyfftw.byte_align(mat, n=fftw_byte_align)
            return numpy_fft.ifftn(mat, shape, axes=axes, overwrite_input=replace, threads=fftw_thread)
    else:
        raise ValueError("invalid FFT library")

    # assign transforms functions
    global FFTN
    global IFFTN
    FFTN = fct_fftn
    IFFTN = fct_ifftn


def get_fft_tensor_keep(mat, replace):
    """
    Get the FFT of a 4D tensor along the first 3D.
    The size of the output is the same of the input
    """

    mat_trf = FFTN(mat, None, (0, 1, 2), replace)

    return mat_trf


def get_fft_tensor_expand(mat, replace):
    """
    Get the FFT of a 4D tensor along the first 3D.
    The size of the output is the double the size of the input
    """

    # get the tensor size
    (nx, ny, nz) = mat.shape[0:3]

    # get the transform
    mat_trf = FFTN(mat, (2*nx, 2*ny, 2*nz), (0, 1, 2), replace)

    return mat_trf


def get_ifft_tensor(mat, replace):
    """
    Get the iFFT of a 4D tensor along the first 3D.
    The size of the output is the same of the input
    """

    mat_trf = IFFTN(mat, None, (0, 1, 2), replace)

    return mat_trf



"""
Module managing the configuration of the program.
    - The config variables are accessed with attributes.
    - First the default configuration is loaded ("pypeec/data/config.yaml").
    - Afterward, a custom file can be loaded via the environment variable ("PYPEEC").

Warning
-------
    - The configuration is loaded by different PyPEEC modules.
    - The configuration cannot be updated after the initialization.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import sys
import os.path
import pathlib
import importlib.resources
from pypeec import io
from pypeec.lib_check import check_data_config


def __getattr__(name):
    """
    Wrapper to access the config data with attributes.
    """

    if name in DATA_CONFIG:
        return DATA_CONFIG[name]
    else:
        return None


class _DictToAttributes:
    """
    Wrapper to access a dictionary with attributes.
    """

    def __init__(self, data):
        """
        Set a dictionary.
        """

        self.data = data

    def __getattr__(self, name):
        """
        Access the dictionary with attributes.
        """

        return self.data[name]


def _parse_config(data_config):
    """
    Make the config accessible with attributes.
    """

    for tag, data_config_tmp in data_config.items():
        if isinstance(data_config_tmp, dict):
            data_config[tag] = _DictToAttributes(data_config_tmp)
        else:
            data_config[tag] = data_config_tmp

    return data_config


def _set_file_config(file_config):
    """
    Load and set a configuration file.
    This function is be called immediately after initializing the module.
    """

    # load the config file
    try:
        # load the config file
        data_config = io.load_input(file_config)

        # check the data integrity and complete the config
        check_data_config.check_data_config(data_config)

        # make the dictionary accessible with attributes
        data_config = _parse_config(data_config)

        # assign config to a global variable
        global DATA_CONFIG
        DATA_CONFIG = data_config

        # assign config to a global variable
        global FILE_CONFIG
        FILE_CONFIG = str(file_config)
    except Exception as ex:
        print("==========================", file=sys.stderr)
        print("INVALID CONFIGURATION FILE", file=sys.stderr)
        print("==========================", file=sys.stderr)
        print(str(ex), file=sys.stderr)
        print("==========================", file=sys.stderr)
        sys.exit(1)


def get_file_config():
    """
    Return the loaded configuration file.

    Returns
    -------
    file_config : string
        String with the loaded configuration file.
    """

    global FILE_CONFIG

    return FILE_CONFIG


def get_data_config():
    """
    Return the loaded configuration data.

    Returns
    -------
    data_config : dict
        Dict with the loaded configuration data.
    """

    global DATA_CONFIG

    return DATA_CONFIG


def get_version():
    """
    Get the PyPEEC version.

    Returns
    -------
    version : string
        String with the PyPEEC version.
    """

    try:
        with importlib.resources.open_text("pypeec.data", "version.txt") as file_version:
            version = file_version.read()
    except FileNotFoundError:
        version = 'x.x.x'

    return version


# init config data
DATA_CONFIG = {}
FILE_CONFIG = None

# load the default config files
with importlib.resources.path("pypeec.data", "config.yaml") as file_config:
    _set_file_config(file_config)

# check for env variables
file_config = os.getenv("PYPEEC")
if file_config is not None:
    file_config = pathlib.Path(file_config)
    _set_file_config(file_config)



"""
Module for the user provided options.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

from pypeec.lib_check import datachecker


def check_tag_list(data_check, tag_list):
    """
    Check the list of plots to be shown.
    """

    # check the list of plots
    datachecker.check_list("tag_list", tag_list, can_be_empty=True, sub_type=str, can_be_none=True)
    if tag_list is not None:
        for tag in tag_list:
            datachecker.check_choice("tag_list", tag, data_check)


def check_plot_options(plot_mode, folder, name):
    """
    Check the plot mode (display or not the plots).
    """

    datachecker.check_choice("plot_mode", plot_mode, ["qt", "nb", "save", "none"])
    datachecker.check_string("name", name, can_be_empty=False, can_be_none=True)
    datachecker.check_folder("folder", folder, can_be_none=True)


def check_data_options(is_truncated):
    """
    Check the plot mode (display or not the plots).
    """

    datachecker.check_boolean("is_truncated", is_truncated)


def check_data_voxel(data_voxel):
    """
    Check the voxel data.
    """

    # extract field
    key_list = ["duration", "is_truncated", "data_geom"]
    datachecker.check_dict("data_voxel", data_voxel, key_list=key_list)

    is_truncated = data_voxel["is_truncated"]
    data_geom = data_voxel["data_geom"]

    # check data
    datachecker.check_dict("data_geom", data_geom)
    datachecker.check_boolean("is_truncated", is_truncated)
    datachecker.check_assert("is_truncated", not is_truncated, "truncated input data cannot be used")

    return data_geom


def check_data_solution(data_solution):
    """
    Check the solution data.
    """

    # extract field
    key_list = ["duration", "is_truncated", "data_init", "data_sweep"]
    datachecker.check_dict("data_solution", data_solution, key_list=key_list)

    is_truncated = data_solution["is_truncated"]
    data_init = data_solution["data_init"]
    data_sweep = data_solution["data_sweep"]

    # check data
    datachecker.check_dict("data_geom", data_init)
    datachecker.check_dict("data_geom", data_sweep)
    datachecker.check_boolean("is_truncated", is_truncated)
    datachecker.check_assert("is_truncated", not is_truncated, "truncated input data cannot be used")

    return data_init, data_sweep



"""
Module for extracting the terminal currents and voltage from a solution.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np


def _get_value_terminal(source, src, sink):
    """
    Get the terminal values (current and voltage).
    """

    V = source[src]["V"]-source[sink]["V"]
    I = (source[src]["I"]-source[sink]["I"])/2

    return V, I


def _get_load_terminal(source, terminal_list):
    """
    Get the terminal currents and voltages for a specific sweep.
    """

    # init list
    V_vec = []
    I_vec = []

    # get the solution
    for terminal in terminal_list:
        # extract the terminal name
        src = terminal["src"]
        sink = terminal["sink"]

        # extract the terminal quantities
        (V, I) = _get_value_terminal(source, src, sink)

        # add the quantities
        V_vec.append(V)
        I_vec.append(I)

    # assign the data
    V_vec = np.array(V_vec, dtype=np.complex_)
    I_vec = np.array(I_vec, dtype=np.complex_)

    return V_vec, I_vec


def get_extract(data_solution, sweep_list, terminal_list):
    """
    Get the terminal currents and voltages for given sweeps and windings.
    """

    # extract the data
    data_init = data_solution["data_init"]
    data_sweep = data_solution["data_sweep"]

    # check status
    assert isinstance(data_init, dict), "invalid solution"
    assert isinstance(data_init, dict), "invalid solution"

    # init data
    V_mat = []
    I_mat = []
    freq_vec = []
    has_converged_vec = []

    # extract data
    for sweep in sweep_list:
        # extract the data
        data_sweep_tmp = data_sweep[sweep]
        freq = data_sweep_tmp["freq"]
        source = data_sweep_tmp["source"]
        has_converged = data_sweep_tmp["has_converged"]

        # compute
        (V_vec, I_vec) = _get_load_terminal(source, terminal_list)

        # assign
        V_mat.append(V_vec)
        I_mat.append(I_vec)
        freq_vec.append(freq)
        has_converged_vec.append(has_converged)

    # assemble frequency
    freq = np.mean(freq_vec)

    # check frequency
    assert np.allclose(freq_vec, freq), "invalid solution: invalid frequency"
    assert np.all(has_converged_vec), "invalid solution: convergence issue"

    # create data
    terminal = {
        "n_solution": len(sweep_list),
        "n_winding": len(terminal_list),
        "V_mat": np.array(V_mat, dtype=np.complex_).transpose(),
        "I_mat": np.array(I_mat, dtype=np.complex_).transpose(),
        "freq": freq,
    }

    return terminal



"""
Module for checking the data type and content.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import os.path
import numpy as np


def check_dict(name, data, key_list=None, sub_type=None, can_be_empty=True, can_be_none=False):
    """
    Check a dict.
    """

    # check if none
    if can_be_none and data is None:
        return

    # check type
    if not isinstance(data, dict):
        raise AssertionError("%s: should be a dict" % name)
    if (not can_be_empty) and (len(data) == 0):
        raise AssertionError("%s: should not be empty" % name)

    # check value
    for tag, value in data.items():
        if not isinstance(tag, str):
            raise AssertionError("%s: all the keys should be strings" % name)
        if len(tag) == 0:
            raise AssertionError("%s: keys cannot be empty" % name)
        if (sub_type is not None) and not isinstance(value, sub_type):
            raise AssertionError("%s: invalid type" % name)
    if key_list is not None:
        for tag in key_list:
            if tag not in data:
                raise AssertionError("%s: dict is incomplete: %s" % (name, tag))


def check_list(name, data, sub_type=None, can_be_empty=True, can_be_none=False):
    """
    Check a list.
    """

    # check if none
    if can_be_none and data is None:
        return

    # check type
    if not isinstance(data, list):
        raise AssertionError("%s: should be a list" % name)
    if (not can_be_empty) and (len(data) == 0):
        raise AssertionError("%s: should not be empty" % name)

    # check value
    for value in data:
        if (sub_type is not None) and not isinstance(value, sub_type):
            raise AssertionError("%s: invalid type" % name)


def check_float_pts(name, data, size=None, is_positive=False, can_be_zero=True, can_be_empty=True, can_be_none=False):
    """
    Check a 2D float array.
    """

    # check if none
    if can_be_none and data is None:
        return

    # cast to array
    data = np.array(data)

    # check type and value
    if (not can_be_empty) and (len(data) == 0):
        raise AssertionError("%s: cannot be empty" % name)
    for data_tmp in data:
        check_float_array(name, data_tmp, size, is_positive, can_be_zero, can_be_empty)


def check_integer_pts(name, data, size=None, is_positive=False, can_be_zero=True, can_be_empty=True, can_be_none=False):
    """
    Check a 2D integer array.
    """

    # check if none
    if can_be_none and data is None:
        return

    # cast to array
    data = np.array(data)

    # check type and value
    if (not can_be_empty) and (len(data) == 0):
        raise AssertionError("%s: cannot be empty" % name)
    for data_tmp in data:
        check_integer_array(name, data_tmp, size, is_positive, can_be_zero, can_be_empty)


def check_float_array(name, data, size=None, is_positive=False, can_be_zero=True, can_be_empty=True, can_be_none=False):
    """
    Check a float array.
    """

    # check if none
    if can_be_none and data is None:
        return

    # cast to array
    data = np.array(data)

    # check type and value
    if len(data.shape) != 1:
        raise AssertionError("%s: invalid shape" % name)
    if size is not None:
        if not (len(data) == size):
            raise AssertionError("%s: invalid array size" % name)
    if (not can_be_empty) and (len(data) == 0):
        raise AssertionError("%s: cannot be empty" % name)
    if len(data) > 0:
        if not np.issubdtype(data.dtype, np.floating):
            raise AssertionError("%s: invalid array type" % name)
        if is_positive and not np.all(data >= 0):
            raise AssertionError("%s: should be positive" % name)
        if (not can_be_zero) and np.any(data == 0):
            raise AssertionError("%s: cannot be zero" % name)


def check_integer_array(name, data, size=None, is_positive=False, can_be_zero=True, can_be_empty=True, can_be_none=False):
    """
    Check an integer array.
    """

    # check if none
    if can_be_none and data is None:
        return

    # cast to array
    data = np.array(data)

    # check type and value
    if len(data.shape) != 1:
        raise AssertionError("%s: invalid shape" % name)
    if size is not None:
        if not (len(data) == size):
            raise AssertionError("%s: invalid array size" % name)
    if (not can_be_empty) and (len(data) == 0):
        raise AssertionError("%s: cannot be empty" % name)
    if len(data) > 0:
        if not np.issubdtype(data.dtype, np.integer):
            raise AssertionError("%s: invalid array type" % name)
        if is_positive and not np.all(data >= 0):
            raise AssertionError("%s: should be positive" % name)
        if (not can_be_zero) and np.any(data == 0):
            raise AssertionError("%s: cannot be zero" % name)


def check_index_array(name, data, size=None, bnd=None, can_be_empty=True, can_be_none=False):
    """
    Check an integer array.
    """

    # check if none
    if can_be_none and data is None:
        return

    # cast to array
    data = np.array(data)

    # check type and value
    if len(data.shape) != 1:
        raise AssertionError("%s: invalid shape" % name)
    if size is not None:
        if not (len(data) == size):
            raise AssertionError("%s: invalid array size" % name)
    if (not can_be_empty) and (len(data) == 0):
        raise AssertionError("%s: cannot be empty" % name)
    if len(data) > 0:
        if not np.issubdtype(data.dtype, np.integer):
            raise AssertionError("%s: invalid array type" % name)
        if not (len(np.unique(data)) == len(data)):
            raise AssertionError("%s: indices should be unique" % name)
        if bnd is not None:
            if not (np.all(data >= 0) and np.all(data < bnd)):
                raise AssertionError("%s: invalid index range" % name)


def check_float(name, data, is_positive=False, can_be_zero=True, can_be_none=False):
    """
    Check a float.
    """

    # check if none
    if can_be_none and data is None:
        return

    # check type
    if not np.issubdtype(type(data), np.floating):
        raise AssertionError("%s: should be a float" % name)

    # check value
    if is_positive and not (data >= 0):
        raise AssertionError("%s: should be positive" % name)
    if (not can_be_zero) and (data == 0):
        raise AssertionError("%s: cannot be zero" % name)


def check_integer(name, data, is_positive=False, can_be_zero=True, can_be_none=False):
    """
    Check a integer.
    """

    # check if none
    if can_be_none and data is None:
        return

    # check type
    if not np.issubdtype(type(data), np.integer):
        raise AssertionError("%s: should be an integer" % name)

    # check value
    if is_positive and not (data >= 0):
        raise AssertionError("%s: should be positive" % name)
    if (not can_be_zero) and (data == 0):
        raise AssertionError("%s: cannot be zero" % name)


def check_boolean(name, data, can_be_none=False):
    """
    Check a boolean.
    """

    # check if none
    if can_be_none and data is None:
        return

    # check type
    if not isinstance(data, bool):
        raise AssertionError("%s: should be a boolean" % name)


def check_string(name, data, can_be_empty=True, can_be_none=False):
    """
    Check a string.
    """

    # check if none
    if can_be_none and data is None:
        return

    # check type
    if not isinstance(data, str):
        raise AssertionError("%s: should be a string" % name)
    if (not can_be_empty) and (len(data) == 0):
        raise AssertionError("%s: cannot be empty" % name)


def check_choice(name, data, choice_list, can_be_none=False):
    """
    Check a string within given choices.
    """

    # check if none
    if can_be_none and data is None:
        return

    # check type
    if not isinstance(data, str):
        raise AssertionError("%s: should be a string" % name)
    if data not in choice_list:
        raise AssertionError("%s: invalid value: %s" % (name, data))


def check_folder(name, folder, can_be_none=False):
    """
    Check that a folder is existing.
    """

    # check if none
    if can_be_none and folder is None:
        return

    # check the type
    if not isinstance(folder, str):
        raise AssertionError("%s: folder name should be a string" % name)

    # check that the folder exist
    if not os.path.isdir(folder):
        raise AssertionError("%s: folder does not exist: %s" % (name, folder))


def check_filename(name, filename, can_be_none=False):
    """
    Check that a filename is existing.
    """

    # check if none
    if can_be_none and filename is None:
        return

    # check the type
    if not isinstance(filename, str):
        raise AssertionError("%s: file name should be a string" % name)

    # check that the file exist
    if not os.path.isfile(filename):
        raise AssertionError("%s: file does not exist: %s" % (name, filename))

    return filename


def check_assert(name, cond, msg):
    """
    Check an assertion.
    """

    if not cond:
        raise AssertionError("%s: %s" % (name, msg))



"""
Test the complete workflow (mesher, viewer, solver, and plotter).

True unit tests are not implemented.
Only non-regression integration tests are implemented.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import os
import unittest
from tests.code import test_pypeec
from tests.code import test_read_write
from tests.code import test_generate


class TestWorkflow(unittest.TestCase):
    """
    Run the complete workflow.
    """

    def _check_mesher(self, mesher, mesher_ref):
        """
        Check the results produced by the mesher.
        """

        # get the results
        n_total_ref = mesher_ref["n_total"]
        n_used_ref = mesher_ref["n_used"]

        # extract the solution
        n_total = mesher["n_total"]
        n_used = mesher["n_used"]

        # check solution
        self.assertEqual(n_total, n_total_ref, msg="invalid number of voxels (complete grid)")
        self.assertEqual(n_used, n_used_ref, msg="invalid number of voxels (non-empty voxels)")

    def _check_solver(self, solver, solver_ref, test_tol):
        """
        Check the results produced by the solver.
        """

        # get the results
        freq_ref = solver_ref["freq"]
        has_converged_ref = solver_ref["has_converged"]
        P_tot_ref = solver_ref["P_tot"]
        W_tot_ref = solver_ref["W_tot"]

        # extract the solution
        freq = solver["freq"]
        has_converged = solver["has_converged"]
        P_tot = solver["P_tot"]
        W_tot = solver["W_tot"]

        # check solution
        self.assertEqual(has_converged, has_converged_ref, msg="invalid convergence")
        self.assertAlmostEqual(freq, freq_ref, delta=test_tol*freq_ref, msg="invalid frequency")
        self.assertAlmostEqual(P_tot, P_tot_ref, delta=test_tol*P_tot_ref, msg="invalid losses")
        self.assertAlmostEqual(W_tot, W_tot_ref, delta=test_tol*W_tot_ref, msg="invalid energy")

    def _check_results(self, mesher, solver, mesher_ref, solver_ref, test_tol):
        """
        Check the results.
        """

        # check the mesher
        self._check_mesher(mesher, mesher_ref)

        # check the solver sweep names
        self.assertEqual(solver.keys(), solver_ref.keys(), "invalid sweep")

        # check the solver results
        for solver_tmp, solver_ref_tmp in zip(solver.values(), solver_ref.values()):
            self._check_solver(solver_tmp, solver_ref_tmp, test_tol)

    def run_test(self, tag, name, use_script):
        """
        Run the workflow and check the results.
        """

        # get env var
        test_tol = os.getenv("TEST_TOL")
        test_check = os.getenv("TEST_CHECK")
        test_set = os.getenv("TEST_SET")

        # check env variables
        self.assertIsNotNone(test_tol, "invalid test env variables")
        self.assertIsNotNone(test_check, "invalid test env variables")
        self.assertIsNotNone(test_set, "invalid test env variables")

        # cast env variables
        test_tol = float(test_tol)
        test_check = bool(int(test_check))
        test_set = bool(int(test_set))

        # generate the results
        (data_voxel, data_solution) = test_pypeec.run_workflow(name, use_script)

        # parse the obtained results
        (mesher, solver) = test_generate.generate_results(data_voxel, data_solution)

        # write the reference results
        if test_set:
            test_read_write.write_results(tag, mesher, solver)

        # load and check the results
        if test_check:
            (mesher_ref, solver_ref) = test_read_write.read_results(tag)
            self._check_results(mesher, solver, mesher_ref, solver_ref, test_tol)


def set_test(test_class, tag, name, use_script):
    """
    Add a test case to the test class.
    """

    # function describing the test
    def get(self):
        return test_class.run_test(self, tag, name, use_script)

    # dynamically add the method as an attribute
    setattr(test_class, "test/" + tag, get)



"""
Different functions for handling creating the PEEC dense matrices:
    - inductance and potential matrix
    - magnetic-electric coupling matrices

Function operators are returned for performing the matrix-vector multiplications.
The multiplication can either be done with the dense matrices or with FFT circulant tensors.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np
import scipy.sparse as sps
from pypeec.lib_matrix import matrix_multiply


def _get_face_voxel_indices(n, idx_v, idx_f, A_net, offset):
    """
    Create a matrix to project a variable into a voxel variable.
    The variable is a vector and a single component (x, y, or z) is projected.
    Create an index vector for the non-zero voxel vector components.
    """

    # extract the voxel data
    (nx, ny, nz) = n
    nv = nx*ny*nz

    # get the local indices (face indices of the incidence matrix)
    idx_local = np.in1d(idx_f, np.arange(offset*nv, (offset+1)*nv, dtype=np.int_))

    # slice matrix (columns)
    A_net = A_net[:, idx_local]

    # get face indices (get the non-zero entry of the projected voxel variable)
    idx_local = A_net.getnnz(axis=1) > 0

    # slice matrix (rows)
    A_net = A_net[idx_local, :]

    # scale the matrix for vector projection
    A_net = 0.5*np.abs(A_net)

    # extract the indices
    idx = idx_v[idx_local]

    # add index offset (such that the x, z, and z indices are not identical)
    idx = offset*nv+idx

    return A_net, idx


def _get_face_voxel_matrix(n, idx_v, idx_f, A_net):
    """
    Create a matrix to project a face vector into a voxel vector.
    All the vector components (x, y, and z) are projected.
    Create an index vector for the non-zero voxel vector components.
    """

    # get the incidence matrix between the faces and voxels
    (A_net_x, idx_x) = _get_face_voxel_indices(n, idx_v, idx_f, A_net, 0)
    (A_net_y, idx_y) = _get_face_voxel_indices(n, idx_v, idx_f, A_net, 1)
    (A_net_z, idx_z) = _get_face_voxel_indices(n, idx_v, idx_f, A_net, 2)

    # assemble incidence matrix and indices
    idx_fv = np.concatenate((idx_x, idx_y, idx_z))
    A_fv_net = sps.block_diag((A_net_x, A_net_y, A_net_z))

    return A_fv_net, idx_fv


def _get_operator_zeros(idx_out):
    """
    Get a linear operator returning zeros.
    """

    # vector with zeros
    var_out = np.zeros(len(idx_out), dtype=np.complex_)

    # function returning zeros
    def op(_):
        return var_out

    return op


def get_inductance_matrix(n, d, idx_f, G_self, G_mutual, has_domain, mult_type):
    """
    Extract the inductance matrix of the system (used for the full system).

    The problem contains n_f internal faces.
    The voxel structure has the following size: (nx, ny, nz).
    The green tensor has the following size: (nx, ny, nz, 1).

    The tensor is then used to create a matrix-vector linear operator:
        - input size: n_f
        - output size: n_f
    """

    # check if the matrix is required
    if not has_domain:
        L = np.zeros(len(idx_f), dtype=np.float_)
        L_op = _get_operator_zeros(idx_f)
        return L, L_op

    # vacuum permeability
    mu = 4*np.pi*1e-7

    # extract the voxel data
    (nx, ny, nz) = n
    (dx, dy, dz) = d
    nv = nx*ny*nz

    # get the direction of the faces (x, y, z)
    idx_fx = np.in1d(idx_f, np.arange(0*nv, 1*nv, dtype=np.int_))
    idx_fy = np.in1d(idx_f, np.arange(1*nv, 2*nv, dtype=np.int_))
    idx_fz = np.in1d(idx_f, np.arange(2*nv, 3*nv, dtype=np.int_))

    # scaling factor
    scale = np.zeros(len(idx_f), dtype=np.complex_)
    scale[idx_fx] = mu/(dy**2*dz**2)
    scale[idx_fy] = mu/(dx**2*dz**2)
    scale[idx_fz] = mu/(dx**2*dy**2)

    # self-inductance for the preconditioner
    L = scale*G_self

    # get the matrix-vector operator
    L_op_tmp = matrix_multiply.get_operator_inductance(idx_f, G_mutual, mult_type)

    # function describing the coupling from the electric to the magnetic faces
    def L_op(var_f):
        res_f = scale*L_op_tmp(var_f)
        return res_f

    return L, L_op


def get_potential_matrix(d, idx_v, G_self, G_mutual, has_domain, mult_type):
    """
    Extract the potential matrix of the system.

    The problem contains n_v non-empty voxels.
    The voxel structure has the following size: (nx, ny, nz).
    The green tensor has the following size: (nx, ny, nz, 1).

    The tensor is then used to create a matrix-vector linear operator:
        - input size: n_v
        - output size: n_v
    """

    # check if the matrix is required
    if not has_domain:
        P = 0.0
        P_op = _get_operator_zeros(idx_v)
        return P, P_op

    # vacuum permeability
    mu = 4*np.pi*1e-7

    # extract the voxel data
    (dx, dy, dz) = d

    # scaling factor
    scale = 1/(mu*dx**2*dy**2*dz**2)

    # self-inductance for the preconditioner
    P = scale*G_self

    # get the matrix-vector operator
    P_op_tmp = matrix_multiply.get_operator_potential(idx_v, G_mutual, mult_type)

    # function describing the coupling from the electric to the magnetic faces
    def P_op(var_v):
        res_v = scale*P_op_tmp(var_v)
        return res_v

    return P, P_op


def get_coupling_matrix(n, idx_vc, idx_vm, idx_fc, idx_fm, A_net_c, A_net_m, K_tsr, has_coupling, mult_type):
    """
    Extract the magnetic-electric coupling matrices.

    The problem contains n_fc internal electric faces.
    The problem contains n_fm internal magnetic faces.
    The voxel structure has the following size: (nx, ny, nz).
    The coupling tensor has the following size: (nx, ny, nz, 3).

    Ideally, the coupling matrix would directly operate between the faces.
    The face to face coupling matrix is not a Toeplitz matrix with Toeplitz blocks.
    The voxel to voxel coupling matrix is a Toeplitz matrix with Toeplitz blocks.
    Therefore, the following procedure is used:
        - the face vector is projected into a voxel vector
        - the multiplication is done with the coupling tensor
        - voxel vector is projected into a face vector

    It should be noted that this projection has a negative impact on the achieved accuracy.
    However, this step is currently required for obtaining Toeplitz matrices.

    For the electric coupling, the matrix-vector linear operator has the following size:
        - input size: n_fm
        - output size: n_fc

    For the magnetic coupling, the matrix-vector linear operator has the following size:
        - input size: n_fc
        - output size: n_fm
    """

    # check if the matrix is required
    if not has_coupling:
        K_op_c = _get_operator_zeros(idx_fc)
        K_op_m = _get_operator_zeros(idx_fm)
        return K_op_c, K_op_m

    # get the face voxel incidence matrix
    (A_fv_net_c, idx_fvc) = _get_face_voxel_matrix(n, idx_vc, idx_fc, A_net_c)
    (A_fv_net_m, idx_fvm) = _get_face_voxel_matrix(n, idx_vm, idx_fm, A_net_m)

    # get the coupling operator (voxel to voxel)
    (K_op_c_tmp, K_op_m_tmp) = matrix_multiply.get_operator_coupling(idx_fvc, idx_fvm, K_tsr, mult_type)

    # function describing the coupling from the magnetic to the electric faces
    def K_op_c(var_fm):
        var_fc = A_fv_net_c.transpose()*K_op_c_tmp(A_fv_net_m*var_fm)
        return var_fc

    # function describing the coupling from the electric to the magnetic faces
    def K_op_m(var_fc):
        var_fm = A_fv_net_m.transpose()*K_op_m_tmp(A_fv_net_c*var_fc)
        return var_fm

    return K_op_c, K_op_m



"""
Add data to the PyVista objects for the plotter:
    - scalar variables
    - vector variables
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np
import numpy.linalg as lna


def set_voxel_material(voxel, idx, idx_vc, idx_vm, idx_src_c, idx_src_v):
    """
    Add the material and source description to the unstructured grid.
    Integers are used to encode the different tags.
    """

    # find position
    idx_vc_local = np.in1d(idx, idx_vc)
    idx_vm_local = np.in1d(idx, idx_vm)
    idx_src_c_local = np.in1d(idx, idx_src_c)
    idx_src_v_local = np.in1d(idx, idx_src_v)
    idx_vcm_local = np.logical_and(idx_vc_local, idx_vm_local)

    # init the material
    material = np.empty(len(idx), dtype=np.int_)

    # assign the materials
    material[idx_vc_local] = 1
    material[idx_vm_local] = 2
    material[idx_vcm_local] = 3

    # assign the sources
    material[idx_src_c_local] = 4
    material[idx_src_v_local] = 5

    # get sorted indices
    idx_sort = np.argsort(idx)

    # sort data
    material = material[idx_sort]

    # assign the data to the geometry
    voxel["material"] = material

    return voxel


def set_voxel_scalar(voxel, idx, idx_var, var, name):
    """
    Add a scalar variable to the unstructured grid (complex variable).
    """

    # find the variable indices
    idx_s = np.argsort(idx)
    idx_p = np.searchsorted(idx, idx_var, sorter=idx_s)
    idx_var_local = idx_s[idx_p]

    # assign scalar variable (nan for the voxels where the variable is not defined)
    var_all = np.full(len(idx), np.nan+1j*np.nan, dtype=np.complex_)
    var_all[idx_var_local] = var

    # sort the variable
    var_all = var_all[idx_s]

    # assign potential
    voxel[name + "_re"] = np.real(var_all)
    voxel[name + "_im"] = np.imag(var_all)
    voxel[name + "_abs"] = np.abs(var_all)

    return voxel


def set_voxel_vector(voxel, idx, idx_var, var, name):
    """
    Add a vector variable to the unstructured grid (complex variable).
    The norm (scalar field) and the direction (vector field) are added.
    """

    # find the variable indices
    idx_s = np.argsort(idx)
    idx_p = np.searchsorted(idx, idx_var, sorter=idx_s)
    idx_var_local = idx_s[idx_p]

    # assign vector variable (nan for the voxels where the variable is not defined)
    var_all = np.full((len(idx), 3), np.nan+1j*np.nan, dtype=np.complex_)
    var_all[idx_var_local] = var

    # sort the variable
    var_all = var_all[idx_s]

    # assign the vector and the norm
    voxel[name + "_vec_re"] = np.real(var_all)
    voxel[name + "_vec_im"] = np.imag(var_all)
    voxel[name + "_norm"] = lna.norm(var_all, axis=1)

    return voxel


def set_magnetic_field(point, H_pts):
    """
    Add the magnetic field to the point cloud.
    The norm (scalar field) and the direction (vector field) are added.
    """

    # assign the vector and the norm
    point["H_vec_re"] = np.real(H_pts)
    point["H_vec_im"] = np.imag(H_pts)
    point["H_norm"] = lna.norm(H_pts, axis=1)

    return point


def get_voxel(idx_vc, idx_vm):
    """
    Get the indices of the non-empty voxels.
    """

    idx_all = np.concatenate((idx_vc, idx_vm))
    idx_all = np.unique(idx_all)

    return idx_all



"""
Module for checking the PNG mesher data.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

from pypeec.lib_check import datachecker


def _check_domain_color(domain_color):
    """
    Check that the mapping between the pixel colors and the domains is valid.
    """

    # check type
    datachecker.check_dict("domain_color", domain_color, can_be_empty=False)

    # check value
    for color_list in domain_color.values():
        datachecker.check_integer_pts("domain_color", color_list, size=4, can_be_empty=False, is_positive=True, can_be_zero=True)


def _check_layer_stack(layer_stack):
    """
    Check the validity of the image layer stack.
    """

    # check type
    datachecker.check_list("layer_stack", layer_stack, can_be_empty=False)

    # check value
    for layer_stack_tmp in layer_stack:
        # check type
        key_list = ["n_layer", "filename_list"]
        datachecker.check_dict("layer_stack", layer_stack_tmp, key_list=key_list)

        # extract the data
        n_layer = layer_stack_tmp["n_layer"]
        filename_list = layer_stack_tmp["filename_list"]

        # check data
        datachecker.check_integer("n_layer", n_layer, is_positive=True, can_be_zero=False)
        datachecker.check_list("filename_list", filename_list, can_be_empty=False, sub_type=str)

        # check files
        for filename in filename_list:
            datachecker.check_filename("filename_list", filename)


def _check_param(param):
    """
    Check the voxel structure parameters.
    """

    # check type
    key_list = ["d", "c", "size"]
    datachecker.check_dict("param", param, key_list=key_list)

    # check data
    datachecker.check_integer_array("size", param["size"], size=2, is_positive=True, can_be_zero=False)
    datachecker.check_float_array("d", param["d"], size=3, is_positive=True, can_be_zero=False)
    datachecker.check_float_array("c", param["c"], size=3, is_positive=False, can_be_zero=True)


def check_data_voxelize(data_voxelize):
    """
    Check the data used for voxelization.
    """

    # check type
    key_list = ["param", "domain_color", "layer_stack"]
    datachecker.check_dict("data_voxelize", data_voxelize, key_list=key_list)

    # extract field
    param = data_voxelize["param"]
    domain_color = data_voxelize["domain_color"]
    layer_stack = data_voxelize["layer_stack"]

    # check voxel structure parameters
    _check_param(param)

    # check domain colors
    _check_domain_color(domain_color)

    # check the PNG layer stack
    _check_layer_stack(layer_stack)

    # get the domain name
    domain_list = domain_color.keys()

    return domain_list



"""
Different functions for handling the specified geometry (materials and sources).
Create and manage the indices of the different voxels and faces.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np
from pypeec import log

# get a logger
LOGGER = log.get_logger("PROBLEM")


def get_material_indices(material_idx):
    """
    Get the indices of the voxels for the different material types.
    """

    # array for the indices
    idx_vc = np.empty(0, dtype=np.int_)
    idx_vm = np.empty(0, dtype=np.int_)

    # populate the arrays
    for tag, material_idx_tmp in material_idx.items():
        # extract the data
        material_type = material_idx_tmp["material_type"]
        idx = material_idx_tmp["idx"]

        # get the electric materials
        if material_type in ["electric", "electromagnetic"]:
            idx_vc = np.append(idx_vc, idx)

        # get the magnetic materials
        if material_type in ["magnetic", "electromagnetic"]:
            idx_vm = np.append(idx_vm, idx)

    return idx_vc, idx_vm


def get_material_pos(material_idx, idx_vc, idx_vm):
    """
    Get the indices of the electric and magnetic materials.
    Convert from voxel indices to solver indices.
    """

    # init material dict
    material_pos = {}

    # parse the material domains
    for tag, material_idx_tmp in material_idx.items():
        # extract the data
        idx = material_idx_tmp["idx"]

        # get the position of the material domain
        idx_vc_tmp = np.in1d(idx_vc, idx)
        idx_vm_tmp = np.in1d(idx_vm, idx)

        # find indices
        idx_vc_tmp = np.flatnonzero(idx_vc_tmp)
        idx_vm_tmp = np.flatnonzero(idx_vm_tmp)

        # assign the losses
        material_pos[tag] = {"idx_vc": idx_vc_tmp, "idx_vm": idx_vm_tmp}

    return material_pos


def get_source_indices(source_idx):
    """
    Get the indices of the source voxels (current and voltage sources).
    """

    # array for the source indices
    idx_src_c = np.empty(0, dtype=np.int_)
    idx_src_v = np.empty(0, dtype=np.int_)

    # populate the arrays with the current sources
    for tag, source_idx_tmp in source_idx.items():
        # extract the data
        source_type = source_idx_tmp["source_type"]
        idx = source_idx_tmp["idx"]

        # assign the indices
        if source_type == "current":
            idx_src_c = np.append(idx_src_c, idx)
        if source_type == "voltage":
            idx_src_v = np.append(idx_src_v, idx)

    return idx_src_c, idx_src_v


def get_source_pos(source_idx, idx_vc, idx_src_c, idx_src_v):
    """
    Get the indices of the source terminal voltages and currents.
    Convert from voxel indices to solver indices.
    """

    # init source dict
    source_pos = {}

    # assemble indices
    idx_src = np.concatenate((idx_src_c, idx_src_v))

    # parse the source terminals
    for tag, source_idx_tmp in source_idx.items():
        # extract the data
        idx = source_idx_tmp["idx"]

        # get the position of the voltage terminals
        idx_vc_tmp = np.in1d(idx_vc, idx)
        idx_src_tmp = np.in1d(idx_src, idx)

        # find indices
        idx_vc_tmp = np.flatnonzero(idx_vc_tmp)
        idx_src_tmp = np.flatnonzero(idx_src_tmp)

        # assign the current and voltage
        source_pos[tag] = {"idx_vc": idx_vc_tmp, "idx_src": idx_src_tmp}

    return source_pos


def get_reduce_matrix(pts_vox, A_vox, idx_v):
    """
    Reduce the matrices to the non-empty voxels and compute face indices.

    The voxel structure has the following size: (nx, ny, nz).
    The problem contains n_v non-empty voxels and n_f internal faces.
    At the input, the complete coordinate matrix is provided: (nx*ny*nz, 3).
    At the input, the complete incidence matrix is provided: (nx*ny*nz, 3*nx*ny*nz).
    At the output, the reduced coordinate matrix is provided: (n_v, 3).
    At the output, the reduced incidence matrix is provided: (n_v, n_f).
    The indices of the internal faces are also computed.
    """

    # reduce the size of the voxel coordinate amtrix
    pts_net = pts_vox[idx_v, :]

    # reduce the size of the incidence matrix (only the non-empty voxels)
    A_net = A_vox[idx_v, :]

    # indices of the all the internal faces (global face indices, 0:3*n)
    idx_f = np.sum(np.abs(A_net), axis=0) == 2
    idx_f = np.flatnonzero(idx_f)

    # reduce the size of the incidence matrix (only the internal faces)
    A_net = A_net[:, idx_f]

    return pts_net, A_net, idx_f


def get_status(n, idx_vc, idx_vm, idx_fc, idx_fm, idx_src_c, idx_src_v):
    """
    Get a dict summarizing the problem size.
    Total number of voxels, number of non-empty voxels, number of faces, and number of sources.
    """

    # extract the voxel data
    (nx, ny, nz) = n

    # count
    n_voxel_total = nx*ny*nz
    n_face_total = 3*nx*ny*nz
    n_voxel_electric = len(idx_vc)
    n_voxel_magnetic = len(idx_vm)
    n_face_electric = len(idx_fc)
    n_face_magnetic = len(idx_fm)
    n_src_current = len(idx_src_c)
    n_src_voltage = len(idx_src_v)

    # fraction of voxels
    n_voxel_used = n_voxel_electric+n_voxel_magnetic
    n_face_used = n_face_electric+n_face_magnetic
    ratio_voxel = n_voxel_used/n_voxel_total
    ratio_face = n_face_used/n_face_total

    # assign data
    problem_status = {
        "n_voxel_total": n_voxel_total,
        "n_face_total": n_face_total,
        "n_voxel_used": n_voxel_used,
        "n_face_used": n_face_used,
        "n_voxel_electric": n_voxel_electric,
        "n_voxel_magnetic": n_voxel_magnetic,
        "n_face_electric": n_face_electric,
        "n_face_magnetic": n_face_magnetic,
        "n_src_current": n_src_current,
        "n_src_voltage": n_src_voltage,
        "ratio_voxel": ratio_voxel,
        "ratio_face": ratio_face,
    }

    # display status
    LOGGER.debug("n_voxel_total = %d" % n_voxel_total)
    LOGGER.debug("n_voxel_used = %d" % n_voxel_used)
    LOGGER.debug("n_face_total = %d" % n_face_total)
    LOGGER.debug("n_face_used = %d" % n_face_used)
    LOGGER.debug("n_voxel_electric = %d" % n_voxel_electric)
    LOGGER.debug("n_voxel_magnetic = %d" % n_voxel_magnetic)
    LOGGER.debug("n_face_electric = %d" % n_face_electric)
    LOGGER.debug("n_face_magnetic = %d" % n_face_magnetic)
    LOGGER.debug("n_src_current = %d" % n_src_current)
    LOGGER.debug("n_src_voltage = %d" % n_src_voltage)
    LOGGER.debug("ratio_voxel = %.2e" % ratio_voxel)
    LOGGER.debug("ratio_face = %.2e" % ratio_face)

    return problem_status



"""
Module for checking the solver problem data.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

from pypeec.lib_check import datachecker


def _check_field(val_dict, key_list, var_type, orientation_type):
    """
    Check a value dict for a variable (lumped / distributed / isotropic / anisotropic).
    """

    # check type
    datachecker.check_dict("val", val_dict, key_list=key_list)

    # check data
    for tag, val in val_dict.items():
        if var_type == "lumped" and orientation_type is None:
            datachecker.check_float(tag, val)
        elif var_type == "lumped" and orientation_type == "isotropic":
            datachecker.check_float(tag, val)
        elif var_type == "lumped" and orientation_type == "anisotropic":
            datachecker.check_float_array(tag, val, size=3)
        elif var_type == "distributed" and orientation_type is None:
            datachecker.check_float_array(tag, val)
        elif var_type == "distributed" and orientation_type == "isotropic":
            datachecker.check_float_array(tag, val)
        elif var_type == "distributed" and orientation_type == "anisotropic":
            datachecker.check_float_pts(tag, val, size=3)


def _check_material_def(material_def):
    """
    Check that the material definition is valid.
    """

    # check type
    datachecker.check_dict("material_def", material_def, can_be_empty=False)

    # check value
    for material_def_tmp in material_def.values():
        # check type
        key_list = ["material_type", "orientation_type", "var_type", "domain_list"]
        datachecker.check_dict("material_def", material_def_tmp, key_list=key_list)

        # extract the data
        var_type = material_def_tmp["var_type"]
        material_type = material_def_tmp["material_type"]
        orientation_type = material_def_tmp["orientation_type"]
        domain_list = material_def_tmp["domain_list"]

        # check data
        var_type_list = ["lumped", "distributed"]
        orientation_type_list = ["isotropic", "anisotropic"]
        material_type_list = ["electric", "magnetic", "electromagnetic"]
        datachecker.check_choice("var_type", var_type, var_type_list)
        datachecker.check_choice("material_type", material_type, material_type_list)
        datachecker.check_choice("orientation_type", orientation_type, orientation_type_list)
        datachecker.check_list("domain_list", domain_list, can_be_empty=False, sub_type=str)


def _check_source_def(source_def):
    """
    Check that the source definition is valid.
    """

    # check type
    datachecker.check_dict("source_def", source_def, can_be_empty=False)

    # check value
    for source_def_tmp in source_def.values():
        # check type
        key_list = ["source_type", "var_type", "domain_list"]
        datachecker.check_dict("source_def", source_def_tmp, key_list=key_list)

        # extract the data
        var_type = source_def_tmp["var_type"]
        source_type = source_def_tmp["source_type"]
        domain_list = source_def_tmp["domain_list"]

        # check data
        source_type_list = ["current", "voltage"]
        var_type_list = ["lumped", "distributed"]
        datachecker.check_choice("var_type", var_type, var_type_list)
        datachecker.check_choice("source_type", source_type, source_type_list)
        datachecker.check_list("domain_list", domain_list, can_be_empty=False, sub_type=str)


def _check_sweep_param(sweep_param, material_def, source_def):
    """
    Check that the excitation definition (materials and sources) is valid.
    """

    # check type
    key_list = ["freq", "material_val", "source_val"]
    datachecker.check_dict("sweep_param", sweep_param, key_list=key_list)

    # extract field
    freq = sweep_param["freq"]
    material_val = sweep_param["material_val"]
    source_val = sweep_param["source_val"]

    # check data
    datachecker.check_float("freq", freq, is_positive=True, can_be_zero=True)

    # check type
    datachecker.check_dict("material_val", material_val, can_be_empty=False)
    datachecker.check_dict("source_val", source_val, can_be_empty=False)

    # check value
    for tag, material_val_tmp in material_val.items():
        # extract the data
        datachecker.check_choice("tag", tag, material_def)
        material_type = material_def[tag]["material_type"]
        orientation_type = material_def[tag]["orientation_type"]
        var_type = material_def[tag]["var_type"]

        # get the source value
        if material_type == "electric":
            key_list = ["rho_re", "rho_im"]
        elif material_type == "magnetic":
            key_list = ["chi_re", "chi_im"]
        elif material_type == "electromagnetic":
            key_list = ["rho_re", "rho_im", "chi_re", "chi_im"]
        else:
            raise ValueError("invalid material type")

        # check variable
        _check_field(material_val_tmp, key_list, var_type, orientation_type)

    # check value
    for tag, source_val_tmp in source_val.items():
        # extract the data
        datachecker.check_choice("tag", tag, source_def)
        source_type = source_def[tag]["source_type"]
        var_type = source_def[tag]["var_type"]

        # get the source value
        if source_type == "current":
            key_list = [
                "I_re", "I_im",
                "Y_re", "Y_im",
            ]
        elif source_type == "voltage":
            key_list = [
                "V_re", "V_im",
                "Z_re", "Z_im",
            ]
        else:
            raise ValueError("invalid source type")

        # check variable
        _check_field(source_val_tmp, key_list, var_type, None)


def check_data_problem(data_problem):
    """
    Check the solver problem data:
        - frequency
        - material definition
        - source definition
    """

    # check type
    key_list = ["material_def", "source_def", "sweep_config", "sweep_param"]
    datachecker.check_dict("data_problem", data_problem, key_list=key_list)

    # extract field
    material_def = data_problem["material_def"]
    source_def = data_problem["source_def"]
    sweep_config = data_problem["sweep_config"]
    sweep_param = data_problem["sweep_param"]

    # check material and source
    _check_material_def(material_def)
    _check_source_def(source_def)

    # check excitation values
    datachecker.check_dict("sweep_param", sweep_param, can_be_empty=False)
    for sweep_param_tmp in sweep_param.values():
        _check_sweep_param(sweep_param_tmp, material_def, source_def)

    # get sweep names
    sweep_list = sweep_param.keys()

    # check the sweep configuration
    datachecker.check_dict("sweep_config", sweep_config, can_be_empty=False)
    for tag_run, tag_init in sweep_config.items():
        datachecker.check_choice("tag_run", tag_run, sweep_list, can_be_none=False)
        datachecker.check_choice("tag_init", tag_init, sweep_list, can_be_none=True)



"""
Module for checking the solver config data.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

from pypeec.lib_check import datachecker


def check_data_config(data_config):
    """
    Check that the config is valid.
    """

    # check type
    key_list = [
        "LOGGING_OPTIONS",
        "PROBLEM_MAX_SIZE",
        "DISPLAY_LOGO",
        "SWEEP_POOL",
        "PAUSE_GUI",
    ]
    datachecker.check_dict("data_config", data_config, key_list=key_list)

    # check logging options
    LOGGING_OPTIONS = data_config["LOGGING_OPTIONS"]
    key_list = [
        "FORMAT",
        "LEVEL",
        "INDENTATION",
        "EXCEPTION_TRACE",
        "USE_COLOR",
        "DEF_COLOR",
    ]
    datachecker.check_dict("LOGGING_OPTIONS", LOGGING_OPTIONS, key_list=key_list)

    # check logging options content
    datachecker.check_dict("FORMAT", LOGGING_OPTIONS["FORMAT"])
    datachecker.check_string("LEVEL", LOGGING_OPTIONS["LEVEL"], can_be_empty=False)
    datachecker.check_integer("INDENTATION", LOGGING_OPTIONS["INDENTATION"], is_positive=True)
    datachecker.check_boolean("EXCEPTION_TRACE", LOGGING_OPTIONS["EXCEPTION_TRACE"])
    datachecker.check_boolean("USE_COLOR", LOGGING_OPTIONS["USE_COLOR"])
    datachecker.check_dict("DEF_COLOR", LOGGING_OPTIONS["DEF_COLOR"])

    # check logging format definition
    key_list = [
        "LOGGER",
        "TIMESTAMP_FMT",
        "DURATION_FMT",
        "TIMESTAMP_TRC",
        "DURATION_TRC",
    ]
    datachecker.check_dict("FORMAT", LOGGING_OPTIONS["FORMAT"], key_list=key_list)
    datachecker.check_string("LOGGER", LOGGING_OPTIONS["FORMAT"]["LOGGER"], can_be_empty=False)
    datachecker.check_string("TIMESTAMP_FMT", LOGGING_OPTIONS["FORMAT"]["TIMESTAMP_FMT"], can_be_empty=False)
    datachecker.check_string("DURATION_FMT", LOGGING_OPTIONS["FORMAT"]["DURATION_FMT"], can_be_empty=False)
    datachecker.check_integer("TIMESTAMP_TRC", LOGGING_OPTIONS["FORMAT"]["TIMESTAMP_TRC"], is_positive=True)
    datachecker.check_integer("DURATION_TRC", LOGGING_OPTIONS["FORMAT"]["DURATION_TRC"], is_positive=True)

    # check logging color definition
    key_list = [
        "CL_DEBUG",
        "CL_INFO",
        "CL_WARNING",
        "CL_ERROR",
        "CL_CRITICAL",
        "CL_RESET",
    ]
    datachecker.check_dict("DEF_COLOR", LOGGING_OPTIONS["DEF_COLOR"], key_list=key_list)
    datachecker.check_string("CL_DEBUG", LOGGING_OPTIONS["DEF_COLOR"]["CL_DEBUG"], can_be_empty=False)
    datachecker.check_string("CL_INFO", LOGGING_OPTIONS["DEF_COLOR"]["CL_INFO"], can_be_empty=False)
    datachecker.check_string("CL_WARNING", LOGGING_OPTIONS["DEF_COLOR"]["CL_WARNING"], can_be_empty=False)
    datachecker.check_string("CL_ERROR", LOGGING_OPTIONS["DEF_COLOR"]["CL_ERROR"], can_be_empty=False)
    datachecker.check_string("CL_CRITICAL", LOGGING_OPTIONS["DEF_COLOR"]["CL_CRITICAL"], can_be_empty=False)
    datachecker.check_string("CL_RESET", LOGGING_OPTIONS["DEF_COLOR"]["CL_RESET"], can_be_empty=False)

    # check problem size limit
    PROBLEM_MAX_SIZE = data_config["PROBLEM_MAX_SIZE"]
    key_list = [
        "VOXEL_TOTAL",
        "VOXEL_USED",
    ]
    datachecker.check_dict("PROBLEM_MAX_SIZE", PROBLEM_MAX_SIZE, key_list=key_list)
    datachecker.check_integer("VOXEL_TOTAL", PROBLEM_MAX_SIZE["VOXEL_TOTAL"], is_positive=True, can_be_zero=False, can_be_none=True)
    datachecker.check_integer("VOXEL_USED", PROBLEM_MAX_SIZE["VOXEL_USED"], is_positive=True, can_be_zero=False, can_be_none=True)

    # check other switches
    datachecker.check_boolean("DISPLAY_LOGO", data_config["DISPLAY_LOGO"])
    datachecker.check_float("PAUSE_GUI", data_config["PAUSE_GUI"], is_positive=True, can_be_zero=True, can_be_none=True)
    datachecker.check_integer("SWEEP_POOL", data_config["SWEEP_POOL"], is_positive=True, can_be_zero=False, can_be_none=True)



"""
Module for factorizing sparse matrix.

This module is used as a common interface for different solvers:
    - SuperLU is typically slower but is always available (integrated with SciPy)
    - UMFPACK is typically faster than SuperLU (available through SciKits)
    - PARDISO is typically faster than UMFPACK (available through Pydiso)
    - PyAMG is typically slow but uses less memory (risk of convergence issues)

Todo
----
    - The warning triggered by UMFPACK should be handled in a cleaner way.
    - Such warnings are triggered by ill-conditioned matrices.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import os
import warnings
from pypeec import log

# get a logger
LOGGER = log.get_logger("FACTOR")

# dummy options
LIBRARY = None
IMPORTLIB = None
PYAMG_OPTIONS = None
PARDISO_OPTIONS = None


def _get_fact_superlu(mat):
    """
    Factorize a matrix with SuperLU.
    """

    # factorize the matrix
    try:
        mat_factor = IMPORTLIB.splu(mat)
    except Warning:
        raise RuntimeError("invalid factorization: SuperLU")

    # matrix solver
    def factor(rhs):
        sol = mat_factor.solve(rhs)
        return sol

    return factor


def _get_fact_pardiso(mat):
    """
    Factorize a matrix with PARDISO.
    """

    # get options
    thread_pardiso = PARDISO_OPTIONS["thread_pardiso"]
    thread_mkl = PARDISO_OPTIONS["thread_mkl"]

    # find the number of threads
    if thread_pardiso is None:
        thread_pardiso = os.cpu_count()
    if thread_mkl is None:
        thread_mkl = os.cpu_count()

    # set number of threads
    IMPORTLIB.set_mkl_pardiso_threads(thread_pardiso)
    IMPORTLIB.set_mkl_threads(thread_mkl)

    # factorize the matrix
    try:
        mat = mat.tocsr()
        mat_factor = IMPORTLIB.MKLPardisoSolver(mat, factor=True, verbose=False)
    except Warning:
        raise RuntimeError("invalid factorization: PARDISO")

    # matrix solver
    def factor(rhs):
        sol = mat_factor.solve(rhs)
        return sol

    return factor


def _get_fact_pyamg(mat):
    """
    Factorize a matrix with PyAMG.
    """

    # get options
    tol = PYAMG_OPTIONS["tol"]
    solver = PYAMG_OPTIONS["solver"]
    krylov = PYAMG_OPTIONS["krylov"]

    # factorize the matrix
    try:
        mat = mat.tocsr()
        if solver == "root":
            solver = IMPORTLIB.rootnode_solver(mat)
        elif solver == "adapt":
            (solver, work) = IMPORTLIB.adaptive_sa_solver(mat)
        else:
            raise ValueError("invalid AMF solver name")
    except Warning:
        raise RuntimeError("invalid factorization: PyAMG")

    # matrix solver
    def factor(rhs):
        sol = solver.solve(rhs, tol=tol, accel=krylov)
        return sol

    return factor


def _get_fact_umfpack(mat):
    """
    Factorize a matrix with UMFPACK.
    """

    # factorize the matrix
    try:
        mat_factor = IMPORTLIB.splu(mat)
    except Warning:
        raise RuntimeError("invalid factorization: UMFPACK")

    # matrix solver
    def factor(rhs):
        sol = mat_factor.solve(rhs)
        return sol

    return factor


def _get_factorize_sub(mat):
    """
    Factorize a sparse matrix (main function).
    """

    # check shape
    nnz = mat.size
    (nx, ny) = mat.shape

    # factorization for empty matrices
    def factor_dummy(rhs):
        return rhs

    # check if the matrix is empty
    if (nx, ny) == (0, 0):
        return factor_dummy

    # compute matrix density
    density = nnz/(nx*ny)

    # display
    LOGGER.debug("matrix size: (%d, %d)" % (nx, ny))
    LOGGER.debug("matrix elements: %d" % nnz)
    LOGGER.debug("matrix density: %.2e" % density)

    # factorize the matrix
    LOGGER.debug("compute factorization")
    if LIBRARY == "SuperLU":
        factor = _get_fact_superlu(mat)
    elif LIBRARY == "UMFPACK":
        factor = _get_fact_umfpack(mat)
    elif LIBRARY == "PARDISO":
        factor = _get_fact_pardiso(mat)
    elif LIBRARY == "PyAMG":
        factor = _get_fact_pyamg(mat)
    else:
        raise ValueError("invalid matrix factorization library")

    # display the status
    LOGGER.debug("factorization success")

    return factor


def set_options(factorization_options):
    """
    Assign the options and load the right libray.
    """

    # assign global variable
    global LIBRARY
    global PYAMG_OPTIONS
    global PARDISO_OPTIONS
    LIBRARY = factorization_options["library"]
    PYAMG_OPTIONS = factorization_options["pyamg_options"]
    PARDISO_OPTIONS = factorization_options["pardiso_options"]

    # import the right library
    if LIBRARY == "SuperLU":
        import scipy.sparse.linalg as lib_tmp
    elif LIBRARY == "UMFPACK":
        import scikits.umfpack as lib_tmp
    elif LIBRARY == "PARDISO":
        import pydiso.mkl_solver as lib_tmp
    elif LIBRARY == "PyAMG":
        import pyamg.aggregation as lib_tmp
    elif LIBRARY == "none":
        lib_tmp = None
    else:
        raise ValueError("invalid factorization library")

    # assign import library to global
    global IMPORTLIB
    IMPORTLIB = lib_tmp

    # prevent problematic matrices to trigger warnings
    warnings.filterwarnings("error", module="scikits.umfpack")
    warnings.filterwarnings("error", module="pydiso.mkl_solver")


def get_factorize(name, mat):
    """
    Factorize a sparse matrix (log wrapper).
    """

    LOGGER.debug("factorization: %s" % name)
    with log.BlockIndent():
        factor = _get_factorize_sub(mat)

    return factor


def get_solve(factor, rhs):
    """
    Solve an equation system with a given factorization.
    """

    # check that the factorization is valid
    if factor is None:
        raise RuntimeError("invalid factorization")

    # solve the equation system
    sol = factor(rhs)

    return sol



"""
Module for managing plotting windows with PyVista and Matplotlib.
The Qt framework or Jupyter are used for showing the plots.

Warning
-------
    - This module is using different more or less dirty hacks.
    - This module is likely to be non-portable across platforms (tested on Linux x64).
    - This module is likely to break with newer/older versions of the dependencies.

Todo
----
    - Making many plots can lead to segmentation fault with PyVista.
    - Not sure if the problem lies with PyPEEC, PyVista, PyVistaQt or Vtk.

Todo
----
    - A delay is adding between the plots when using the Qt framework.
    - This is a dirty workaround for a race condition in PyVista/PyVistaQt.

Todo
----
    - Quitting or making screenshot might cause a crash of PyVista.
    - The reason is that the rendering is not forced properly with multiple windows.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import os
import sys
import ctypes
import signal
import time
import os.path
import importlib.resources
import pyvista
import pyvistaqt
import matplotlib
import matplotlib.pyplot
import PyQt5.QtWidgets
import PyQt5.QtGui
from pypeec import log
from pypeec import config

# get a logger
LOGGER = log.get_logger("PLOTGUI")

# get config
PAUSE_GUI = config.PAUSE_GUI


class PlotGui:
    """
    Manage PyVista and Matplotlib plots.
    Three different plot mode are available:
        - "qt", the Qt framework is used for the rendering (default).
        - "nb", the plots are rendered within the Jupyter notebook.
        - "save", the plots are not shown but saved as screenshots.
        - "none", the plots are not shown (test mode).
    """

    def __init__(self, plot_mode, folder, name):
        """
        Constructor.
        Init the plots.
        """

        # assign variable
        self.plot_mode = plot_mode
        self.folder = folder
        self.name = name
        self.pl_list = []
        self.fig_list = []

        # create the Qt App
        if self.plot_mode == "qt":
            self.app = PyQt5.QtWidgets.QApplication([])
        else:
            self.app = None

        # set the app ID in order to get a consistent icon on MS Windows
        if (self.plot_mode == "qt") and (os.name == "nt"):
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID("pypeec")

        # setup PyVista and Matplotlib
        if self.plot_mode == "qt":
            pyvista.set_plot_theme("default")
            matplotlib.use("QtAgg")
        if self.plot_mode == "nb":
            pyvista.set_plot_theme("default")
            pyvista.set_jupyter_backend("trame")

    @staticmethod
    def _get_plotter_pyvista_qt(title, show_menu, window_size):
        """
        Create a PyVista plotter for the Qt framework.
        """

        # cast window size
        if window_size is not None:
            window_size = tuple(window_size)

        # get Qt plotter
        pl = pyvistaqt.BackgroundPlotter(
            show=False,
            toolbar=show_menu,
            menu_bar=show_menu,
            title=title,
            window_size=window_size,
        )
        # set icon
        with importlib.resources.path("pypeec.data", "pypeec.png") as file_icon:
            pl.set_icon(str(file_icon))

        # pause to avoid race conditions
        if PAUSE_GUI is not None:
            time.sleep(PAUSE_GUI)

        return pl

    @staticmethod
    def _get_plotter_pyvista_nb(notebook_size):
        """
        Create a PyVista plotter for Jupyter notebooks.
        """

        # cast window size
        if notebook_size is not None:
            notebook_size = tuple(notebook_size)

        # create plotter
        pl = pyvista.Plotter(notebook=True, window_size=notebook_size)

        return pl

    @staticmethod
    def _get_plotter_pyvista_nop(image_size):
        """
        Create a PyVista plotter for silent rendering.
        """

        # cast window size
        if image_size is not None:
            image_size = tuple(image_size)

        # create plotter
        pl = pyvista.Plotter(off_screen=True, window_size=image_size)

        return pl

    @staticmethod
    def _get_figure_matplotlib_qt(title, show_menu, window_size):
        """
        Create a Matplotlib figure for the Qt framework.
        """

        # get the icon
        with importlib.resources.path("pypeec.data", "pypeec.png") as file_icon:
            res_icon = PyQt5.QtGui.QIcon(str(file_icon))

        # get the figure
        fig = matplotlib.pyplot.figure(tight_layout=True)

        # set the Qt options
        man = matplotlib.pyplot.get_current_fig_manager()
        man.window.setWindowTitle(title)
        man.window.setWindowIcon(res_icon)

        # set the menu
        if show_menu:
            man.toolbar.show()
        else:
            man.toolbar.hide()

        # set window size
        if window_size is not None:
            (sx, sy) = window_size
            man.window.resize(sx, sy)

        # pause to avoid race conditions
        if PAUSE_GUI is not None:
            time.sleep(PAUSE_GUI)

        return fig

    @staticmethod
    def _get_figure_matplotlib_nb(notebook_size):
        """
        Create a Matplotlib figure for Jupyter notebooks.
        """

        # create figure
        fig = matplotlib.pyplot.figure(tight_layout=True)

        # set window size
        if notebook_size is not None:
            (sx, sy) = notebook_size
            fig.set_size_inches(sx/fig.dpi, sy/fig.dpi)

        return fig

    @staticmethod
    def _get_figure_matplotlib_nop(image_size):
        """
        Create a Matplotlib figure for silent rendering.
        """

        # create figure
        fig = matplotlib.pyplot.figure(tight_layout=True)

        # set window size
        if image_size is not None:
            (sx, sy) = image_size
            fig.set_size_inches(sx/fig.dpi, sy/fig.dpi)

        return fig

    def _show_figure_qt(self):
        """
        Show all the figures (Qt framework).
        """

        # signal handler for closing all the windows
        def signal_handler(*_):
            for tag, pl in self.pl_list:
                pl.app_window.close()
            matplotlib.pyplot.close("all")
            sys.exit(130)

        # signal for quitting the event loop with interrupt signal
        signal.signal(signal.SIGINT, signal_handler)

        # pause to avoid race conditions
        if PAUSE_GUI is not None:
            time.sleep(PAUSE_GUI)

        # show the different PyVista plots
        for tag, pl in self.pl_list:
            pl.app_window.show()

        # show the different Matplotlib plots
        matplotlib.pyplot.show(block=False)

        # enter the event loop
        exit_code = self.app.exec_()

        # check status
        if exit_code != 0:
            RuntimeError("error during the Qt event loop / exit_code = %d" % exit_code)

    def _show_figure_nb(self):
        """
        Show all the figures (Jupyter notebooks)
        """

        # show the different PyVista plots
        for tag, pl in self.pl_list:
            pl.show(jupyter_backend='trame')

        # show the different Matplotlib plots
        matplotlib.pyplot.show(block=False)

    def _show_figure_nop(self):
        """
        Show all the figures (silent rendering)
        """

        # close the different PyVista plots
        for tag, pl in self.pl_list:
            pl.close()

        # close the different Matplotlib plots
        matplotlib.pyplot.close("all")

    def _get_filename(self, tag):
        """
        Get the plot filenames.
        """

        filename = "%s.png" % tag
        if self.name:
            filename = "%s_%s" % (self.name, filename)
        if self.folder:
            filename = os.path.join(self.folder, filename)

        return filename

    def _save_screenshot(self):
        """
        Save all the plots as images.
        """

        # save the PyVista plots
        for tag, pl in self.pl_list:
            filename = self._get_filename(tag)
            pl.screenshot(filename)

        # save the Matplotlib plots
        for tag, fig in self.fig_list:
            filename = self._get_filename(tag)
            fig.savefig(filename)

    def open_pyvista(self, tag, title, data_window):
        """
        Get a PyVista plotter.
        """

        # extract the data
        show_menu = data_window["show_menu"]
        image_size = data_window["image_size"]
        window_size = data_window["window_size"]
        notebook_size = data_window["notebook_size"]

        # prepend tag
        title = tag + " / " + title

        # create the figure
        if self.plot_mode == "qt":
            pl = self._get_plotter_pyvista_qt(title, show_menu, window_size)
        elif self.plot_mode == "nb":
            pl = self._get_plotter_pyvista_nb(notebook_size)
        elif self.plot_mode in ["save", "none"]:
            pl = self._get_plotter_pyvista_nop(image_size)
        else:
            raise ValueError("invalid plot mode")

        # add the plotter to the list
        self.pl_list.append((tag, pl))

        return pl

    def open_matplotlib(self, tag, title, data_window):
        """
        Get a Matplotlib figure.
        """

        # extract the data
        show_menu = data_window["show_menu"]
        image_size = data_window["image_size"]
        window_size = data_window["window_size"]
        notebook_size = data_window["notebook_size"]

        # prepend tag
        title = tag + " / " + title

        # create the figure
        if self.plot_mode == "qt":
            fig = self._get_figure_matplotlib_qt(title, show_menu, window_size)
        elif self.plot_mode == "nb":
            fig = self._get_figure_matplotlib_nb(notebook_size)
        elif self.plot_mode in ["save", "none"]:
            fig = self._get_figure_matplotlib_nop(image_size)
        else:
            raise ValueError("invalid plot mode")

        # add the figure to the list
        self.fig_list.append((tag, fig))

        return fig

    def show(self):
        """
        Show the plots.
        The following behavior is done for the different plot mode:
            - qt: show plot windows with the Qt framework (blocking call)
            - nb: show the plot inside a Jupyter notebook (non-blocking call)
            - nop: close all the plots without showing them (non-blocking call)
        """

        LOGGER.debug("number of PyVista plots: %s" % len(self.pl_list))
        LOGGER.debug("number of Matplotlib plots: %s" % len(self.fig_list))

        if (len(self.pl_list) == 0) and (len(self.fig_list) == 0):
            return

        if self.plot_mode == "qt":
            LOGGER.debug("entering the plot event loop")
            self._show_figure_qt()
        elif self.plot_mode == "nb":
            LOGGER.debug("display notebook plots")
            self._show_figure_nb()
        elif self.plot_mode == "save":
            LOGGER.debug("save and close all the plots")
            self._save_screenshot()
            self._show_figure_nop()
        elif self.plot_mode == "none":
            LOGGER.debug("close all the plots")
            self._show_figure_nop()
        else:
            raise ValueError("invalid plot mode")



"""
Add data to the PyVista objects for the viewer:
    - domain description
    - connected component
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np


def _get_graph_component(idx, connection_def):
    """
    Find the connected components for a specific domain.
    Assign a different scalar for each connected component.
    """

    # init the data with invalid data
    if len(connection_def) == 0:
        tag = np.ones(len(idx), dtype=np.int_)
    else:
        tag = np.zeros(len(idx), dtype=np.int_)

    # find to corresponding connected components
    for i, idx_graph in enumerate(connection_def):
        # find which indices are part of the connected component
        idx_ok = np.in1d(idx, idx_graph)

        # assign the component number to the corresponding indices
        tag[idx_ok] = i+1

    # check that everything was assigned
    if not np.all(tag):
        raise RuntimeError("invalid graph: some voxels are not part of the graph")

    return tag


def _get_geometry_tag(domain_def, connection_def):
    """
    Assign a different integer for each domain.
    Assign a different integer for each connected component.
    """

    # init
    domain = np.empty(0, dtype=np.int_)
    connection = np.empty(0, dtype=np.int_)

    # get the indices and colors
    counter = 1
    for tag, idx_tmp in domain_def.items():
        # assign the color (n different integer for each domain)
        domain_tmp = np.full(len(idx_tmp), counter, dtype=np.int_)

        # find the connected components corresponding to the indices
        connection_tmp = _get_graph_component(idx_tmp, connection_def)

        # append the indices and colors
        domain = np.append(domain, domain_tmp)
        connection = np.append(connection, connection_tmp)

        # update the domain counter
        counter += 1

    return domain, connection


def set_data(voxel, idx, domain_def, connection_def):
    """
    Add the domain and connected component description to the unstructured grid.
    Integers are used to encode the different tags.
    """

    # get the data
    (domain, connection) = _get_geometry_tag(domain_def, connection_def)

    # get sorted indices
    idx = np.argsort(idx)

    # sort data
    domain = domain[idx]
    connection = connection[idx]

    # assign the data to the geometry
    voxel["domain"] = domain
    voxel["connection"] = connection

    return voxel


def get_voxel(domain_def):
    """
    Get the indices of the non-empty voxels.
    """

    idx = np.empty(0, dtype=np.int_)
    for idx_tmp in domain_def.values():
        idx = np.append(idx, idx_tmp)

    return idx



"""
Combine the voxel and problem data into a new solver data structure.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np
from pypeec.lib_check import datachecker


def _check_indices(n, idx_c, idx_m, idx_s):
    """
    Check that the material and source indices are valid.
    The indices should be unique and compatible with the voxel size.
    The source indices should be included in the electric indices.
    Check that there is at least one source per connected electrical.
    """

    # extract the voxel data
    (nx, ny, nz) = n
    nv = nx*ny*nz

    # check the indices
    datachecker.check_index_array("index electric", idx_c, bnd=nv, can_be_empty=False)
    datachecker.check_index_array("index magnetic", idx_m, bnd=nv, can_be_empty=True)
    datachecker.check_index_array("index source", idx_s, bnd=nv, can_be_empty=False)

    # check that the terminal indices are electric indices
    cond = np.all(np.in1d(idx_s, idx_c))
    datachecker.check_assert("index", cond, "source indices should overlap with electric indices")


def _check_source_graph(idx_c, idx_s, connection_def):
    """
    Check that there is at least one source per connected component.
    A connected components without a source would lead to a singular problem.
    """

    for idx_graph in connection_def:
        has_electric = len(np.intersect1d(idx_graph, idx_c)) > 0
        has_source = len(np.intersect1d(idx_graph, idx_s)) > 0

        cond = (not has_electric) or has_source
        datachecker.check_assert("index", cond, "electric components should include at least one source")


def _get_field(val_dict, idx, var_type, orientation_type):
    """
    Cast and check the material and source vectors.
    If the variable is a scalar, cast to an array.
    If the variable is an array, check the length.
    """

    for tag, val in val_dict.items():
        # cast
        if var_type == "lumped" and orientation_type is None:
            val = np.full(len(idx), val, dtype=np.float_)
        elif var_type == "lumped" and orientation_type == "isotropic":
            val = np.full((len(idx), 3), val, dtype=np.float_)
        elif var_type == "lumped" and orientation_type == "anisotropic":
            val = np.array(val, dtype=np.float_)
            val = np.tile(val, (len(idx), 1))
        elif var_type == "distributed" and orientation_type is None:
            val = np.array(val, dtype=np.float_)
        elif var_type == "distributed" and orientation_type == "isotropic":
            val = np.array(val, dtype=np.float_)
            val = np.tile(val, (3, 1)).transpose()
        elif var_type == "distributed" and orientation_type == "anisotropic":
            val = np.array(val, dtype=np.float_)

        # check
        cond = len(val) == len(idx)
        datachecker.check_assert(tag, cond, "vector length does not match the number of voxels")

        # update
        val_dict[tag] = val

    return val_dict


def _get_domain_indices(domain_list, domain_def):
    """
    Get indices from a list of domain names.
    """

    idx_all = np.empty(0, dtype=np.int_)
    for tag in domain_list:
        idx_tmp = np.array(domain_def[tag], dtype=np.int_)
        idx_all = np.append(idx_all, idx_tmp)

    return idx_all


def _get_material_idx(material_def, domain_def):
    """
    Get the indices of the material.
    Create a new dict with the indices in place of the domain names.
    Split the electric and magnetic materials.
    """

    # init
    material_idx = {}
    idx_c = np.empty(0, dtype=np.int_)
    idx_m = np.empty(0, dtype=np.int_)
    domain_cm = []

    for tag, material_def_tmp in material_def.items():
        # extract the data
        var_type = material_def_tmp["var_type"]
        material_type = material_def_tmp["material_type"]
        orientation_type = material_def_tmp["orientation_type"]
        domain_list = material_def_tmp["domain_list"]

        # get indices
        idx = _get_domain_indices(domain_list, domain_def)

        # append domain names
        domain_cm += domain_list

        # assign the indices
        if material_type == "electric":
            idx_c = np.append(idx_c, idx)
        elif material_type == "magnetic":
            idx_m = np.append(idx_m, idx)
        elif material_type == "electromagnetic":
            idx_c = np.append(idx_c, idx)
            idx_m = np.append(idx_m, idx)
        else:
            raise ValueError("invalid material type")

        # assign the material
        material_idx[tag] = {
            "idx": idx,
            "material_type": material_type,
            "var_type": var_type,
            "orientation_type": orientation_type,
        }

    return domain_cm, idx_c, idx_m, material_idx


def _get_source_idx(source_def, domain_def):
    """
    Get the indices of the sources.
    Create a new dict with the indices in place of the domain names.
    """

    # init
    source_idx = {}
    idx_s = np.empty(0, dtype=np.int_)
    domain_s = []

    for tag, source_def_tmp in source_def.items():
        # extract the data
        var_type = source_def_tmp["var_type"]
        source_type = source_def_tmp["source_type"]
        domain_list = source_def_tmp["domain_list"]

        # get indices
        idx = _get_domain_indices(domain_list, domain_def)

        # append indices
        idx_s = np.append(idx_s, idx)

        # append domain names
        domain_s += domain_list

        # get the source value
        source_idx[tag] = {
            "idx": idx,
            "source_type": source_type,
            "var_type": var_type,
        }

    return domain_s, idx_s, source_idx


def _get_sweep_param(sweep_param, material_idx, source_idx):
    """
    Check the size of the material and source values.
    Convert the values to arrays.
    """

    # extract field
    freq = sweep_param["freq"]
    material_val = sweep_param["material_val"]
    source_val = sweep_param["source_val"]

    # check the material domain name
    for tag in material_idx:
        datachecker.check_choice("material domain name", tag, material_val)
    for tag in source_idx:
        datachecker.check_choice("source domain name", tag, source_val)

    # update values
    for tag, material_val_tmp in material_val.items():
        # extract the data
        orientation_type = material_idx[tag]["orientation_type"]
        var_type = material_idx[tag]["var_type"]
        idx = material_idx[tag]["idx"]

        # check type
        material_val[tag] = _get_field(material_val_tmp, idx, var_type, orientation_type)

    # update values
    for tag, source_val_tmp in source_val.items():
        # extract the data
        var_type = source_idx[tag]["var_type"]
        idx = source_idx[tag]["idx"]

        # check type
        source_val[tag] = _get_field(source_val_tmp, idx, var_type, None)

    # assign results
    sweep_param = {"freq": freq, "material_val": material_val, "source_val": source_val}

    return sweep_param


def get_data_solver(data_geom, data_problem, data_tolerance):
    """
    Combine the voxel data, the problem data, and the tolerance data.
    The voxel data contains the mapping between domain names and indices.
    The problem data contains domain names used for the materials and sources.
    The tolerance data contains numerical options.
    The new dict contains the indices used for the materials and sources.
    """

    # extract field
    material_def = data_problem["material_def"]
    source_def = data_problem["source_def"]
    sweep_config = data_problem["sweep_config"]
    sweep_param = data_problem["sweep_param"]

    # extract geometry
    n = data_geom["n"]
    d = data_geom["d"]
    c = data_geom["c"]
    domain_def = data_geom["domain_def"]
    connection_def = data_geom["connection_def"]
    pts_cloud = data_geom["pts_cloud"]

    # get material and source indices
    (domain_cm, idx_c, idx_m, material_idx) = _get_material_idx(material_def, domain_def)
    (domain_s, idx_s, source_idx) = _get_source_idx(source_def, domain_def)

    # check the domain name
    for tag in domain_cm:
        datachecker.check_choice("material domain name", tag, domain_def)
    for tag in domain_s:
        datachecker.check_choice("source domain name", tag, domain_def)

    # get source and material values
    for tag, sweep_param_tmp in sweep_param.items():
        sweep_param[tag] = _get_sweep_param(sweep_param_tmp, material_idx, source_idx)

    # check voxel indices
    _check_indices(n, idx_c, idx_m, idx_s)
    _check_source_graph(idx_c, idx_s, connection_def)

    # check the existence of magnetic domains
    has_electric = len(idx_c) > 0
    has_magnetic = len(idx_m) > 0
    has_coupling = has_electric and has_magnetic

    # assign combined data
    data_solver = {
        "n": n,
        "d": d,
        "c": c,
        "material_idx": material_idx,
        "source_idx": source_idx,
        "has_electric": has_electric,
        "has_magnetic": has_magnetic,
        "has_coupling": has_coupling,
        "pts_cloud": pts_cloud,
    }

    # add tolerance data
    data_solver = {**data_solver, **data_tolerance}

    return data_solver, sweep_config, sweep_param



"""
Module for checking the solver tolerance data.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

from pypeec.lib_check import datachecker


def _check_fft_options(fft_options):
    """
    Check the FFT algorithm options.
    """

    # check type
    key_list = [
        "matrix_split",
        "library",
        "scipy_worker",
        "fftw_thread",
        "fftw_timeout",
        "fftw_byte_align",
    ]
    datachecker.check_dict("fft_options", fft_options, key_list=key_list)

    # extract field
    matrix_split = fft_options["matrix_split"]
    library = fft_options["library"]
    scipy_worker = fft_options["scipy_worker"]
    fftw_thread = fft_options["fftw_thread"]
    fftw_timeout = fft_options["fftw_timeout"]
    fftw_byte_align = fft_options["fftw_byte_align"]

    # check the data
    datachecker.check_boolean("matrix_split", matrix_split)
    datachecker.check_choice("library", library, ["SciPy", "FFTW", "CuPy"])
    datachecker.check_integer("scipy_worker", scipy_worker, is_positive=True, can_be_zero=False, can_be_none=True)
    datachecker.check_integer("fftw_thread", fftw_thread, is_positive=True, can_be_zero=False, can_be_none=True)
    datachecker.check_integer("fftw_byte_align", fftw_byte_align, is_positive=True, can_be_zero=False, can_be_none=True)
    datachecker.check_float("fftw_timeout", fftw_timeout, is_positive=True, can_be_zero=True, can_be_none=True)


def _check_factorization_options(factorization_options):
    """
    Check the matrix factorization options.
    """

    # check type
    key_list = ["library", "pyamg_options", "pardiso_options"]
    datachecker.check_dict("factorization_options", factorization_options, key_list=key_list)

    # extract field
    library = factorization_options["library"]
    pyamg_options = factorization_options["pyamg_options"]
    pardiso_options = factorization_options["pardiso_options"]

    # check the data
    datachecker.check_choice("library", library, ["SuperLU", "UMFPACK", "PARDISO", "PyAMG"])

    # check the data
    key_list = ["tol", "solver", "krylov"]
    datachecker.check_dict("pyamg_options", pyamg_options, key_list=key_list)
    datachecker.check_choice("solver", pyamg_options["solver"], ["adapt", "root"])
    datachecker.check_choice("krylov", pyamg_options["krylov"], ["gmres", "fgmres"], can_be_none=True)
    datachecker.check_float("tol", pyamg_options["tol"], is_positive=True, can_be_zero=False)

    # check the data
    key_list = ["thread_pardiso", "thread_mkl"]
    datachecker.check_dict("pardiso_options", pardiso_options, key_list=key_list)
    datachecker.check_integer("thread_pardiso", pardiso_options["thread_pardiso"], is_positive=True, can_be_zero=False, can_be_none=True)
    datachecker.check_integer("thread_mkl",  pardiso_options["thread_mkl"], is_positive=True, can_be_zero=False, can_be_none=True)


def _check_iter_options(iter_options):
    """
    Check the iterative solver options.
    """

    # check type
    key_list = ["solver", "rel_tol", "abs_tol", "n_inner", "n_outer"]
    datachecker.check_dict("iter_options", iter_options, key_list=key_list)

    # check data
    datachecker.check_choice("solver", iter_options["solver"], ["gmres", "gcrot"])
    datachecker.check_float("rel_tol", iter_options["rel_tol"], is_positive=True, can_be_zero=False)
    datachecker.check_float("abs_tol", iter_options["abs_tol"], is_positive=True, can_be_zero=False)
    datachecker.check_integer("n_inner", iter_options["n_inner"], is_positive=True, can_be_zero=False)
    datachecker.check_integer("n_outer", iter_options["n_outer"], is_positive=True, can_be_zero=False)


def _check_solver_options(solver_options):
    """
    Check the equation system solver options.
    """

    # check type
    key_list = ["coupling", "power_options", "tolerance_options", "segregated_options", "direct_options"]
    datachecker.check_dict("solver_options", solver_options, key_list=key_list)

    # extract field
    coupling = solver_options["coupling"]
    power_options = solver_options["power_options"]
    tolerance_options = solver_options["tolerance_options"]
    segregated_options = solver_options["segregated_options"]
    direct_options = solver_options["direct_options"]

    # check the coupling
    datachecker.check_choice("coupling", coupling, ["direct", "segregated"])

    # check the power options
    key_list = ["stop", "n_min", "n_cmp", "rel_tol", "abs_tol"]
    datachecker.check_dict("power_options", power_options, key_list=key_list)
    datachecker.check_integer("n_min", power_options["n_min"], is_positive=True, can_be_zero=False)
    datachecker.check_integer("n_cmp", power_options["n_cmp"], is_positive=True, can_be_zero=False)
    datachecker.check_float("rel_tol", power_options["rel_tol"], is_positive=True, can_be_zero=False)
    datachecker.check_float("abs_tol", power_options["abs_tol"], is_positive=True, can_be_zero=False)
    datachecker.check_boolean("stop", power_options["stop"])

    # check the tolerance options
    key_list = ["ignore_status", "ignore_res", "rel_tol", "abs_tol"]
    datachecker.check_dict("tolerance_options", tolerance_options, key_list=key_list)
    datachecker.check_float("rel_tol", tolerance_options["rel_tol"], is_positive=True, can_be_zero=False)
    datachecker.check_float("abs_tol", tolerance_options["abs_tol"], is_positive=True, can_be_zero=False)
    datachecker.check_boolean("ignore_status", tolerance_options["ignore_status"])
    datachecker.check_boolean("ignore_res", tolerance_options["ignore_res"])

    # check the direct solver
    _check_iter_options(direct_options)

    # check the segregated solver
    key_list = [
        "rel_tol", "abs_tol",
        "n_min", "n_max",
        "relax_electric", "relax_magnetic",
        "iter_electric_options", "iter_magnetic_options",
    ]
    datachecker.check_dict("segregated_options", segregated_options, key_list=key_list)
    datachecker.check_float("rel_tol", segregated_options["rel_tol"], is_positive=True, can_be_zero=False)
    datachecker.check_float("abs_tol", segregated_options["abs_tol"], is_positive=True, can_be_zero=False)
    datachecker.check_integer("n_min", segregated_options["n_min"], is_positive=True, can_be_zero=False)
    datachecker.check_integer("n_max", segregated_options["n_max"], is_positive=True, can_be_zero=False)
    datachecker.check_float("relax_electric", segregated_options["relax_electric"], is_positive=True, can_be_zero=False)
    datachecker.check_float("relax_magnetic", segregated_options["relax_magnetic"], is_positive=True, can_be_zero=False)
    _check_iter_options(segregated_options["iter_electric_options"])
    _check_iter_options(segregated_options["iter_magnetic_options"])


def _check_condition_options(condition_options):
    """
    Check the matrix condition number check options.
    """

    # check type
    key_list = ["check", "tolerance_electric", "tolerance_magnetic", "norm_options"]
    datachecker.check_dict("condition_options", condition_options, key_list=key_list)

    # extract field
    check = condition_options["check"]
    tolerance_electric = condition_options["tolerance_electric"]
    tolerance_magnetic = condition_options["tolerance_magnetic"]
    norm_options = condition_options["norm_options"]

    # check the data
    datachecker.check_boolean("check", check)
    datachecker.check_float("tolerance_electric", tolerance_electric, is_positive=True, can_be_zero=False)
    datachecker.check_float("tolerance_magnetic", tolerance_magnetic, is_positive=True, can_be_zero=False)

    # check the data
    key_list = ["t_accuracy", "n_iter_max"]
    datachecker.check_dict("norm_options", norm_options, key_list=key_list)
    datachecker.check_integer("t_accuracy", norm_options["t_accuracy"], is_positive=True, can_be_zero=False)
    datachecker.check_integer("n_iter_max", norm_options["n_iter_max"], is_positive=True, can_be_zero=False)


def check_data_tolerance(data_tolerance):
    """
    Check the solver tolerance data:
        - Green functions simplification
        - Coupling functions simplification
        - matrix multiplication algorithm
        - FFT algorithm options
        - matrix factorization options
        - iterative solver options
        - matrix condition options
    """

    # check type
    key_list = [
        "green_simplify",
        "coupling_simplify",
        "mult_type",
        "factorization_options",
        "fft_options",
        "solver_options",
        "condition_options",
    ]
    datachecker.check_dict("data_tolerance", data_tolerance, key_list=key_list)

    # extract field
    mult_type = data_tolerance["mult_type"]
    green_simplify = data_tolerance["green_simplify"]
    coupling_simplify = data_tolerance["coupling_simplify"]

    # check data
    datachecker.check_choice("mult_type", mult_type, ["fft", "direct"])
    datachecker.check_float("green_simplify", green_simplify, is_positive=True, can_be_zero=False)
    datachecker.check_float("coupling_simplify", coupling_simplify, is_positive=True, can_be_zero=False)

    # extract field
    fft_options = data_tolerance["fft_options"]
    factorization_options = data_tolerance["factorization_options"]
    solver_options = data_tolerance["solver_options"]
    condition_options = data_tolerance["condition_options"]

    # check solver and condition check options
    _check_fft_options(fft_options)
    _check_factorization_options(factorization_options)
    _check_solver_options(solver_options)
    _check_condition_options(condition_options)



"""
Module for checking the STL mesher data.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

from pypeec.lib_check import datachecker


def _check_domain_stl(domain_stl):
    """
    Check the validity of the domain definition.
    """

    # check type
    datachecker.check_dict("domain_stl", domain_stl, can_be_empty=False)

    # check content
    for tag, domain_stl_tmp in domain_stl.items():
        datachecker.check_list("domain_stl", domain_stl_tmp, can_be_empty=False)
        for domain_stl_tmp_tmp in domain_stl_tmp:
            # check type
            key_list = ["scale", "offset", "filename"]
            datachecker.check_dict("domain_stl", domain_stl_tmp_tmp, key_list=key_list)

            # extract the data
            scale = domain_stl_tmp_tmp["scale"]
            offset = domain_stl_tmp_tmp["offset"]
            filename = domain_stl_tmp_tmp["filename"]

            # check data
            datachecker.check_float("scale", scale, is_positive=True, can_be_zero=False)
            datachecker.check_float_array("offset", offset, size=3)
            datachecker.check_filename("filename", filename)


def _check_param(param):
    """
    Check the voxel structure parameters.
    """

    # check type
    key_list = ["d", "xyz_min", "xyz_max"]
    datachecker.check_dict("param", param, key_list=key_list)

    # check data
    datachecker.check_float_array("d", param["d"], size=3, is_positive=True, can_be_zero=False)
    datachecker.check_float_array("xyz_min", param["xyz_min"], size=3, is_positive=False, can_be_zero=True, can_be_none=True)
    datachecker.check_float_array("xyz_max", param["xyz_max"], size=3, is_positive=False, can_be_zero=True, can_be_none=True)


def check_data_voxelize(data_voxelize):
    """
    Check the data used for voxelization.
    """

    # check type
    key_list = ["param", "domain_stl"]
    datachecker.check_dict("data_voxelize", data_voxelize, key_list=key_list)

    # extract field
    param = data_voxelize["param"]
    domain_stl = data_voxelize["domain_stl"]

    # check voxel structure parameters
    _check_param(param)

    # check the STL file
    _check_domain_stl(domain_stl)

    # get the domain name
    domain_list = domain_stl.keys()

    return domain_list
